"""GQ Electronics GMC-SE series Geiger Counter Devices."""

from ..connection import Connection
from .device_rfc1201 import DeviceRFC1201


class GMCSE(DeviceRFC1201):
    """GMC-SE"""

    def __init__(
        self,
        port,
        baudrate=115200,
        timeout=5,
        connection=None,
    ):
        """
        Represent a GMC-SE device.

        Parameters
        ----------
        port: None | str
            Exact port (device dev path / com port) e.g. '/dev/ttyUSB0'
            If None, a Connection object is required.
        baudrate: int
            Device baudrate. Default value is the best-known value for the device.
        timeout: int
            Time limit for pyserial to raise timeout.
        connection : pygmc.Connection
            An initialized pygmc connection interface to the USB device.
            Overrides port & baudrate.
        """
        if isinstance(connection, Connection):
            super().__init__(connection)
        elif port and isinstance(baudrate, int):
            conn = Connection(port=port, baudrate=baudrate, timeout=timeout)
            super().__init__(conn)
        else:
            raise ConnectionError(f"Unable to connect port={port} baudrate={baudrate}")

        # https://www.gqelectronicsllc.com/support/GMC_Selection_Guide.htm
        self._flash_memory_size_bytes = 2**21  # 2 MiB
        self._baudrate = baudrate



import pathlib

import pygmc


def test_pygmc_path():
    """Sanity check - we're testing the dev package i.e. not some pygmc in our PATH"""
    pygmc_path = pathlib.Path(pygmc.__file__).parent.parent
    test_path = pathlib.Path(__file__).parent.parent
    assert pygmc_path == test_path



import logging
import re

from .device_rfc1201 import DeviceRFC1201
from .device_rfc1801 import DeviceRFC1801
from .device_spec404 import DeviceSpec404
from .gmc300 import GMC300, GMC300S, GMC300EPlus
from .gmc320 import GMC320, GMC320Plus, GMC320PlusV5
from .gmc500 import GMC500, GMC500Plus
from .gmc600 import GMC600, GMC600Plus
from .gmc800 import GMC800
from .gmcse import GMCSE

logger = logging.getLogger("pygmc.device")


# Match regex to device. Order matters. First match is used.
device_match_list = [
    # GMC300's
    {
        "match_regex": r"GMC-300S",
        "device_class": GMC300S,
        "protocol_class": DeviceRFC1201,
        "version_example": "GMC-300SRe 1.14",
    },
    {
        # Not sure...
        "match_regex": r"GMC-300E\+",
        "device_class": GMC300EPlus,
        "protocol_class": DeviceRFC1201,
        # Open Issue on PyGMC github... make a PR... Provide an example
        "version_example": "",
    },
    {
        # Does this model exist?
        "match_regex": r"GMC-300",
        "device_class": GMC300,
        "protocol_class": DeviceRFC1201,
        # Open Issue on PyGMC github... make a PR... Provide an example
        "version_example": "",
    },
    # GMC320's
    {
        "match_regex": r"GMC-320\+V5",
        "device_class": GMC320PlusV5,
        "protocol_class": DeviceRFC1201,
        # Open Issue on PyGMC github... make a PR... Provide an example
        "version_example": "",
    },
    {
        "match_regex": r"GMC-320\+",
        "device_class": GMC320Plus,
        "protocol_class": DeviceRFC1201,
        # My 2019 GC says GMC320 Plus however <GETVER>> returns 'GMC-320Re 4.26'
        # For sake of passing unittests (and hoping a newer rev adds the '+')
        # Faking version below from "GMC-320Re 4.26" --> "GMC-320+Re 4.26"
        "version_example": "GMC-320+Re 4.26",
    },
    {
        "match_regex": r"GMC-320",
        "device_class": GMC320,
        "protocol_class": DeviceRFC1201,
        # My 2019 GC says GMC320 Plus however <GETVER>> doesn't.
        "version_example": "GMC-320Re 4.26",
    },
    # GMC500's
    {
        "match_regex": r"GMC-500\+",
        "device_class": GMC500Plus,
        "protocol_class": DeviceRFC1801,
        "version_example": "GMC-500+Re 2.22",
    },
    {
        "match_regex": r"GMC-500",
        "device_class": GMC500,
        "protocol_class": DeviceRFC1801,
        # Open Issue on PyGMC github... make a PR... Provide an example
        "version_example": "",
    },
    # GMC600's
    {
        "match_regex": r"GMC-600\+",
        "device_class": GMC600Plus,
        "protocol_class": DeviceRFC1801,
        "version_example": "GMC-600+Re 2.52",  # Thomaz device
    },
    {
        "match_regex": r"GMC-600",
        "device_class": GMC600,
        "protocol_class": DeviceRFC1801,
        # Open Issue on PyGMC github... make a PR... Provide an example
        "version_example": "",
    },
    # GMC800's
    # Listed as RFC1201 in https://www.gqelectronicsllc.com/GMC-800UserGuide.pdf
    # See: https://www.gqelectronicsllc.com/forum/topic.asp?TOPIC_ID=10394
    # PyGMC invented a spec - DeviceSpec404 - For GMC800 with incorrect official spec
    {
        "match_regex": r"GMC-800",
        "device_class": GMC800,
        "protocol_class": DeviceSpec404,
        "version_example": "GMC-800Re1.08",  # Thomaz owned device example
    },
    # GMCSE
    {
        "match_regex": r"GMC-SE",
        "device_class": GMCSE,
        "protocol_class": DeviceRFC1201,
        "version_example": "GMC-SE Re 1.05",
    },
]


def _get_matched_class_from_version(version: str):
    for device_re in device_match_list:
        pattern = device_re["match_regex"]
        m = re.match(pattern=pattern, string=version)
        if m:
            logger.debug(f"pattern={pattern} matched version={version}")
            return device_re["device_class"]

    logger.warning(f"Unable to auto assign Device={version}")
    issues_url = "https://github.com/Wikilicious/pygmc/issues"
    logger.warning(f"Please open an issue {issues_url}")


def auto_get_device_from_discovery_details(device_details):
    """
    Auto get device class from discovery DeviceDetails named tuple.

    Parameters
    ----------
    device_details: namedtuple
        Discovery device details.

    Returns
    -------
    pygmc.devices.BaseDevice | None

    """
    device_class = _get_matched_class_from_version(device_details.version)
    return device_class


def auto_get_device_from_connection(connection):
    """
    Auto get device class.

    Given a connection, use device's version response to identify the best matching
    device class.

    Parameters
    ----------
    connection: pygmc.connection.Connection

    Returns
    -------
    pygmc.devices.BaseDevice | None

    """
    version = connection.get_at_least(b"<GETVER>>", size=7, wait_sleep=0.01).decode()
    logger.debug(f"Device={version}")

    device_class = _get_matched_class_from_version(version)
    return device_class



"""
Mock the connection class to deterministically return results recorded from GMC-320+.
Structured in a novice friendly manner.
"""

import datetime

import pytest

from pygmc import devices

from .mocks import MockConnection

cmd_response_map = {
    b"<GETVER>>": b"GMC-320Re 4.26",
    b"<GETSERIAL>>": b"\xf4\x88\x00g\x1cB\xc2",
    b"<GETCPM>>": b"\x01\xb6",
    b"<GETGYRO>>": b"\xff\x04\x00\x10\x00L\xaa",
    b"<GETVOLT>>": b"*",
    b"<GETDATETIME>>": b"\x17\x0b\n\x10.6\xaa",
    b"<GETCFG>>": b"\x00\x00\x00\x01\x1e\x01\x00d\x00<\x14\xae\xc7>\x00\xf0\x14\xae\xc7?\x03\xe8\x00\x00\xd0@\x05\x00\x00\x00?\x00\x01\x02\x00\x00\x00\x00\xff\xff\xff\xff\xff\xff\x00\x01\x00x\x15\x01\xbc<\x00\x08\xff\x01\x00\xfe\n\x00\x01\n\x00<\x00\x14\xae\xc7>\x17\x0b\n\x10.6\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff",  # noqa
}

device_result_map = {
    "get_version": "GMC-320Re 4.26",
    "get_serial": "f48800671c42c2",
    "get_cpm": 438,
    "get_gyro": (-252, 16, 76),
    "get_voltage": 4.2,
    "get_datetime": datetime.datetime(2023, 11, 10, 16, 46, 54),
    "get_usv_h": 2.8469999576869767,
    "get_config": {
        "Power": 0,
        "Alarm": 0,
        "Speaker": 0,
        "CalibrationCPM_0": 60,
        "CalibrationCPM_1": 240,
        "CalibrationCPM_2": 1000,
        "SaveDataType": 1,
        "MaxCPM": 444,
        "Baudrate": 254,
        "BatteryType": 0,
        "ThresholdMode": 0,
        "ThresholdCPM": 60,
        "Calibration_uSv_0": 0.38999998569488525,
        "Calibration_uSv_1": 1.559999942779541,
        "Calibration_uSv_2": 6.5,
        "IdleTextState": 5,
        "AlarmValue_uSv": 0.5,
        "Threshold_uSv": 0.38999998569488525,
    },
}
parametrize_data = [(k, v) for k, v in device_result_map.items()]


# Mock the connection to a device
# Uses recorded responses device
mock_connection = MockConnection(cmd_response_map)


# Use our fake/mock connection in our real device class
mock_device = devices.DeviceRFC1201(mock_connection)


@pytest.mark.parametrize("cmd,expected", parametrize_data)
def test_expected_results(cmd, expected):
    """
    Test that the method/cmd in the device actually returns the expected value in the device_result_map dict.
    @pytest.mark.parametrize just calls this function for each item in the dictionary.
    """
    # e.g. getattr(mock_device, 'get_cpm') --> mock_device.get_cpm
    result = getattr(mock_device, cmd)()
    assert result == expected
    print(f"{cmd=}")
    print(f"{expected=}")
    print(f"{result=}")


def test_auto_device():
    device = devices.auto_get_device_from_connection(mock_connection)
    assert isinstance(type(device), type(devices.GMC320))


def test_reset_buffers(capfd):
    mock_device.get_config()
    out, err = capfd.readouterr()
    assert "reset_buffers" in out



import sys
import tempfile

import pytest

import pygmc

from .data import data_history_parser
from .mocks import MockConnection

# How to create cmd input for test:
# start_s = struct.pack(">I", 0)[1:]  # 0=start position
# size_s = struct.pack(">H", 110)  # 110=end position
# cmd = b"<SPIR" + start_s + size_s + b">>"
# and set flash-mem read size to 110
cmd_response_map = {
    b"<SPIR\x00\x00\x00\x00n>>": data_history_parser.raw_history_with_notes1,
    b"<SPIR\x00\x00\x00\x00\xd3>>": data_history_parser.raw_history_with_notes2,
    b"<SPIR\x00\x00\x00\x00\n>>": b"\xff" * 10,  # break reading when page is all \xff
}

# device get_history includes columns as first row
tidy_result = data_history_parser.raw_history_with_notes1_tidy.copy()
# The insert below is bad practice and gave me a headache - it mutates the value for
# the next thing to use it. Added .copy()
tidy_result.insert(
    0, ["datetime", "count", "unit", "mode", "reference_datetime", "notes"]
)

# Expect same data for mock_device2
device_result_map = {
    "get_history_data": tidy_result,
}

device_result_map3 = {
    "get_history_data": [
        ["datetime", "count", "unit", "mode", "reference_datetime", "notes"],
    ],
}


# Mock the connection to a device
# Uses recorded responses device
mock_connection = MockConnection(cmd_response_map)


# Use our fake/mock connection in our real device class
mock_device = pygmc.devices.BaseDevice(mock_connection)
mock_device2 = pygmc.devices.BaseDevice(mock_connection)
mock_device3 = pygmc.devices.BaseDevice(mock_connection)

# Override mem/read size
mock_device._flash_memory_size_bytes = 110
mock_device._flash_memory_page_size_bytes = 110

# tests exiting after 100 counts of 255
mock_device2._flash_memory_size_bytes = 110 + 101
mock_device2._flash_memory_page_size_bytes = 110 + 101

# Tests exit pagination condition
mock_device3._flash_memory_size_bytes = 100
mock_device3._flash_memory_page_size_bytes = 10


parametrize_data = [(mock_device, k, v) for k, v in device_result_map.items()]
parametrize_data.extend([(mock_device2, k, v) for k, v in device_result_map.items()])
parametrize_data.extend([(mock_device3, k, v) for k, v in device_result_map3.items()])


@pytest.mark.parametrize("mock_dev,cmd,expected", parametrize_data)
def test_expected_results(mock_dev, cmd, expected):
    """
    Test that the method/cmd in the device actually returns the expected value in the
    device_result_map dict.
    @pytest.mark.parametrize just calls this function for each item in the dictionary.
    """
    # e.g. getattr(mock_dev, 'get_cpm') --> mock_dev.get_cpm
    result = getattr(mock_dev, cmd)()
    print(f"expected-len={len(expected)} | result-len={len(result)}")
    print(result)
    assert result == expected
    print(f"{cmd=}")
    print(f"{expected=}")
    print(f"{result=}")


@pytest.mark.skipif(sys.platform == "win32", reason="does not run on windows")
def test_history_parser_with_file():
    # TypeError: expected str, bytes or os.PathLike object, not _TemporaryFileWrapper
    # Ugh, windows
    f = tempfile.TemporaryFile()
    f.write(b"\xff" * 101)
    h = pygmc.HistoryParser(filename=f)
    assert h.get_data() == []


@pytest.mark.skipif(sys.platform == "win32", reason="does not run on windows")
def test_history_parser_with_file_path():
    # PermissionError on Windows
    f = tempfile.NamedTemporaryFile()
    f.write(b"\xff" * 1)
    fn = f.name
    h = pygmc.HistoryParser(filename=fn)
    assert h.get_data() == []



"""
Mock the connection class to deterministically return results recorded from GMC-500+.
Structured in a novice friendly manner.
"""

import datetime
import getpass

import pytest

from pygmc import devices

from .mocks import MockConnection

cmd_response_map = {
    b"<GETVER>>": b"GMC-500+Re 2.22",
    b"<GETSERIAL>>": b"00!W!W\xf6",
    b"<GETCPM>>": b"\x00\x00\x04\xba",
    b"<GETCPS>>": b"\x00\x00\x00\x13",
    b"<GETMAXCPS>>": b'\x00\x00\x00"',
    b"<GETCPMH>>": b"\x00\x00\x00\x05",
    b"<GETCPML>>": b"\x00\x00\x05\xdc",
    b"<GETGYRO>>": b"\xff\xf9\xff\x0e\x00%\xaa",
    b"<GETVOLT>>": b"4.0v\x00",
    b"<GETDATETIME>>": b"\x17\x0b\n\x12!\x04\xaa",
    b"<GETCFG>>": b"\x00\x00\x00\x00\x1f\x00\x00d\x00d?&ffu0CC\x00\x00\x00\x19@\x9b33\x00?\x00\x00\x00\x00\x02\x03\x00\x00\x00\x00\xff\xff\xff\xff\xff\xff\x00\x01\x00x\n\x05\xe1<\x00\n\xff\x00\x00\x00\n\x00\x01\n\x00d\x00?\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00wangshaofei\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00log2.asp\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00www.gmcmap.com\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00log2.asp\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x02\x00\x02\x03\x00@<\x02\x00\x00\x01\x00x\x00\xc8\x002\x00d\x05\x01\x01\xa2\xa1\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x17\x0b\n\x12!\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff",  # noqa
}

device_result_map = {
    "get_version": "GMC-500+Re 2.22",
    "get_serial": "303021572157f6",
    "get_cpm": 1210,
    "get_cps": 19,
    "get_max_cps": 34,
    "get_cpmh": 5,
    "get_cpml": 1500,
    "get_gyro": (-7, -242, 37),
    "get_voltage": 4.0,
    "get_datetime": datetime.datetime(2023, 11, 10, 18, 33, 4),
    "get_usv_h": 7.864999977043251,
    "get_config": {
        "Power": 0,
        "Alarm": 0,
        "Speaker": 0,
        "CalibrationCPM_0": 100,
        "CalibrationCPM_1": 30000,
        "CalibrationCPM_2": 25,
        "SaveDataType": 2,
        "MaxCPM": 1505,
        "Baudrate": 0,
        "BatteryType": 0,
        "ThresholdMode": 0,
        "ThresholdCPM": 100,
        "Calibration_uSv_0": 0.6499999761581421,
        "Calibration_uSv_1": 195.0,
        "Calibration_uSv_2": 4.849999904632568,
        "IdleTextState": 0,
        "AlarmValue_uSv": 0.5,
        "Threshold_uSv": 0.5,
    },
}
parametrize_data = [(k, v) for k, v in device_result_map.items()]


# Mock the connection to a device
# Uses recorded responses device
mock_connection = MockConnection(cmd_response_map)


# Use our fake/mock connection in our real device class
mock_device = devices.DeviceRFC1801(mock_connection)


@pytest.mark.parametrize("cmd,expected", parametrize_data)
def test_expected_results(cmd, expected):
    """
    Test that the method/cmd in the device actually returns the expected value in the
    device_result_map dict.
    @pytest.mark.parametrize just calls this function for each item in the dictionary.
    """
    # e.g. getattr(mock_device, 'get_cpm') --> mock_device.get_cpm
    result = getattr(mock_device, cmd)()
    assert result == expected
    print(f"{cmd=}")
    print(f"{expected=}")
    print(f"{result=}")


def test_auto_device():
    device = devices.auto_get_device_from_connection(mock_connection)
    assert isinstance(type(device), type(devices.GMC500))


def test_reset_buffers(capfd):
    mock_device.get_config()
    out, err = capfd.readouterr()
    assert "reset_buffers" in out


# uSv Tests
config_calib0 = {
    "CalibrationCPM_0": 100,
    "CalibrationCPM_1": 200,
    "CalibrationCPM_2": 300,
    "Calibration_uSv_0": 1,
    "Calibration_uSv_1": 10,
    "Calibration_uSv_2": 100,
}

# Tests edge case of non-increasing config cpm
config_calib1 = {
    "CalibrationCPM_0": 100,
    "CalibrationCPM_1": 25,
    "CalibrationCPM_2": 300,
    "Calibration_uSv_0": 1,
    "Calibration_uSv_1": 55,
    "Calibration_uSv_2": 100,
}

# Tests edge case of last config not being the largest (GMC-500+ default is like this)
config_calib2 = {
    "CalibrationCPM_0": 100,
    "CalibrationCPM_1": 30000,
    "CalibrationCPM_2": 25,
    "Calibration_uSv_0": 0.6499999761581421,
    "Calibration_uSv_1": 195.0,
    "Calibration_uSv_2": 4.849999904632568,
}

parametrize_calib_data = [
    # config, cpm, expected_result
    (config_calib0, 10, 0.1),
    (config_calib0, 100, 1),
    (config_calib0, 101, 1.0899999999999999),
    (config_calib0, 150, 5.5),
    (config_calib0, 200, 10),
    (config_calib0, 201, 10.900000000000006),
    (config_calib0, 250, 55),
    (config_calib0, 299, 99.10000000000002),
    (config_calib0, 300, 100),
    (config_calib0, 350, 145),  # tests extrapolation
    # edge cases
    (config_calib1, 25, 0.25),
    (config_calib1, 99, 0.99),
    (config_calib1, 101, 67.43636363636364),
    (config_calib1, 250, 91.81818181818181),
    (config_calib1, 300, 100),
    (config_calib1, 350, 108.18181818181819),
    # more edge cases
    (config_calib2, 50, 0.32499998807907104),
    (config_calib2, 1210, 7.864999977043251),
    (config_calib2, 20000, 129.99999999202615),
    (config_calib2, 30000, 195),
    (config_calib2, 35000, 227.50000000398694),
]


@pytest.mark.parametrize("calib_config,cpm,expected", parametrize_calib_data)
def test_usv(calib_config, cpm, expected):
    cmd_response_map_usv = {}
    mock_connection_usv = MockConnection(cmd_response_map_usv)

    # Use our fake/mock connection in our real device class
    mock_device_usv = devices.DeviceRFC1801(mock_connection_usv)

    mock_device_usv._config = calib_config

    print(f"{cpm=} {expected=} | {calib_config=}")

    mock_device_usv.get_cpm = lambda: cpm
    usv = mock_device_usv.get_usv_h()
    usv_from_input = mock_device_usv.get_usv_h(cpm=cpm)

    assert usv == expected, "uSv does not match expected config calibration"
    assert usv == usv_from_input, "manual input cpm does not match equivalent get_cpm"


def test_getpass_password_input(monkeypatch):
    cmd_response_map_usv = {b"<SETWIFIPWDawkins>>": b"\xaa"}
    mock_connection_pswd = MockConnection(cmd_response_map_usv)

    monkeypatch.setattr(getpass, "getpass", lambda: "Dawkins")

    mock_device_pswd = devices.DeviceRFC1801(mock_connection_pswd)
    mock_device_pswd.set_wifi_password()  # empty input to test our monkeypatch getpass

    cmd_called_count = mock_connection_pswd.get_cmd_calls(b"<SETWIFIPWDawkins>>")

    assert cmd_called_count == 1



import logging

from .auto_get_device import (
    auto_get_device_from_connection,
    auto_get_device_from_discovery_details,
)
from .device import BaseDevice
from .device_rfc1201 import DeviceRFC1201
from .device_rfc1801 import DeviceRFC1801
from .device_spec404 import DeviceSpec404
from .gmc300 import GMC300, GMC300S, GMC300EPlus
from .gmc320 import GMC320, GMC320Plus, GMC320PlusV5
from .gmc500 import GMC500, GMC500Plus
from .gmc600 import GMC600, GMC600Plus
from .gmc800 import GMC800
from .gmcse import GMCSE

logger = logging.getLogger("pygmc.device")



import datetime
import logging
import struct
from io import BufferedIOBase

logger = logging.getLogger(__name__)


class HistoryParser:
    """Parse GQ GMC device history data."""

    def __init__(self, data=None, filename=None):
        """
        Parse GMC flash memory saved history data.

        Parameters
        ----------
        data: bytes | None
            Input raw bytes of history
            If left None, filename must be provided.
        filename: str | BufferedIOBase | None
            Path to file to open or an BufferedIOBase i.e. open(file, 'rb')
            data takes priority over filename.

        """
        if isinstance(data, bytes):
            self._raw = _BinFile(data)
        elif isinstance(filename, BufferedIOBase):
            self._raw = filename
        elif filename:
            self._raw = open(filename, "rb")
        else:
            raise TypeError

        self._datetime = None
        self._unit = None
        self._mode = None
        self._columns = [
            "datetime",
            "count",
            "unit",
            "mode",
            "reference_datetime",
            "notes",
        ]
        self._data = []
        # notes = [(datetime, notes), ...]
        self._notes = []
        # context = [(reference_datetime, unit, mode)]
        self._context_history = []
        # reference time
        self._last_reference_datetime = None
        # last notes - add to next data input and reset to blank
        self._last_note = None
        self._eof_check = 0

        # parse
        self._parse()

    def _get_count_data(self, com_str):
        if len(com_str) == 1:
            value = struct.unpack(">B", com_str)[0]
        elif len(com_str) == 2:
            value = struct.unpack(">H", com_str)[0]
        elif len(com_str) == 3:
            # 3 byte data... 2**24 = 16,777,216 max
            value = int.from_bytes(com_str, "big")
            # value = struct.unpack(">I", com_str)[0]
        elif len(com_str) == 4:
            # 4 byte data... really?! Ok, sure
            value = int.from_bytes(com_str, "big")
        else:
            msg = "Unexpected len for count data: len={}".format(len(com_str))
            raise ValueError(msg)

        # we may be at end of file...
        # treat several consecutive 255 as end of file
        if value == 255:
            self._eof_check += 1
        else:
            self._eof_check = 0
        return value

    def _get_context(self, data):
        save_mode = data[8]
        if save_mode == 0:
            unit = "OFF"
            mode_str = "off"
        elif save_mode == 1:
            unit = "CPS"
            mode_str = "every second"
        elif save_mode == 2:
            unit = "CPM"
            mode_str = "every minute"
        elif save_mode == 3:
            unit = "CPM"
            mode_str = "every hour"
        elif save_mode == 4:
            unit = "CPS"
            mode_str = "every second - threshold"
        elif save_mode == 5:
            unit = "CPM"
            mode_str = "every minute - threshold"
        else:
            msg = "Unknown save mode: {}".format(save_mode)
            logger.error(msg)
            return None, "Unknown", "Unknown"

        year = int("20{0:2d}".format(data[0]))
        month = int("{0:2d}".format(data[1]))
        day = int("{0:2d}".format(data[2]))
        hour = int("{0:2d}".format(data[3]))
        minute = int("{0:2d}".format(data[4]))
        second = int("{0:2d}".format(data[5]))
        ref_dt = datetime.datetime(year, month, day, hour, minute, second)
        self._last_reference_datetime = ref_dt
        logger.debug(
            "Context: unit={} mode={} reference_time={}".format(unit, mode_str, ref_dt)
        )
        return ref_dt, unit, mode_str

    def _set_context(self, ref_dt, unit, mode):
        self._datetime = ref_dt
        self._unit = unit
        self._mode = mode
        self._context_history.append((ref_dt, unit, mode))

    def _add_to_df(self, count):
        # It appears that a ref time is dumped out then the next data entry
        # is put in after <mode> time so we add a time delta before recording
        # the entry in the df
        # BUT... it's useful to be able to match context timestamp to df!!! change?
        if self._datetime is None:
            # add count data with other fields as None?
            return
        if "threshold" in self._mode:
            # really hope GMC outputs reference time for every threshold output
            # TODO: investigate.
            pass
        if "second" in self._mode:
            self._datetime += datetime.timedelta(seconds=1)
        elif "minute" in self._mode:
            self._datetime += datetime.timedelta(minutes=1)
        elif "hour" in self._mode:
            self._datetime += datetime.timedelta(hours=1)

        # header=["datetime", "count", "unit", "mode", "reference_datetime", "note"]
        if self._last_note:
            note = self._last_note
            # i.e. add note to very next count data
            self._last_note = None
        else:
            note = None

        data = (
            self._datetime,
            count,
            self._unit,
            self._mode,
            self._last_reference_datetime,
            note,
        )
        self._data.append(data)

        # check suspicious 255 values
        # Ok, Hubert Farnsworth, Rick Sanchez...
        # Only you two are capable of making 100 contiguous counts of 255
        if self._eof_check > 100:
            logger.debug("EOFError raised - 100 contiguous counts of 255")
            raise EOFError

    def _add_notes(self, note):
        # um... guessing here
        try:
            note = note.decode("utf8")
            self._notes.append((self._datetime, note))
            self._last_note = note
        except UnicodeDecodeError:
            if self._eof_check > 5:
                logger.warning(
                    "Possible end of file... unable to decode note: {}".format(note)
                )

    def _parse(self):
        """The core parser flow."""
        # The meat of the matter...
        try:
            while True:
                # command
                com_str = self._raw.read(1)
                if len(com_str) == 0:
                    # nothing left to read...
                    # use raise as exit ship...
                    raise EOFError  # noqa

                com = ord(com_str)
                # 0x55 (85) Could be context, 2-byte count, notes, OR a regular count
                if com == 85:
                    com_str2 = self._raw.read(1)
                    com2 = ord(com_str2)
                    # 0xaa (170)
                    if com2 == 170:
                        com_str3 = self._raw.read(1)
                        com3 = ord(com_str3)
                        # 0x00 (0)
                        # context data - save mode & datetime
                        if com3 == 0:
                            data = self._raw.read(9)
                            ref_dt, unit, mode = self._get_context(data)
                            self._set_context(ref_dt, unit, mode)
                        # 0x01 (1)
                        elif com3 == 1:
                            # two byte count number
                            # so... max CPM is 65,535?
                            data = self._raw.read(2)
                            n = self._get_count_data(data)
                            self._add_to_df(n)
                        # 0x02 (2)
                        elif com3 == 2:
                            # Notes flag
                            # bytes size of notes (so max notes size is 255?)
                            data = self._raw.read(1)
                            size = ord(data)
                            notes = self._raw.read(size)
                            self._add_notes(notes)
                        elif com3 == 3:
                            # three byte count number
                            # max CPM 16,777,216 (including 0) 2^24
                            data = self._raw.read(3)
                            n = self._get_count_data(data)
                            self._add_to_df(n)
                        elif com3 == 4:
                            # four byte count number? 2^32
                            # The last number you'll see! Guaranteed!
                            data = self._raw.read(4)
                            n = self._get_count_data(data)
                            self._add_to_df(n)
                        elif com3 == 5:
                            # tube selection: 0=both
                            tube = self._raw.read(1)
                            # My GMC-500+ had instances where it failed to record tube
                            if tube == b"U":
                                # An unknown bug... start of cmd w/o tube specified
                                # was expecting \x00 (both) or 1 or 2
                                i = self._raw.tell()  # current read position
                                self._raw.seek(i - 1)  # go back one byte
                                # have to read more than 1 byte to get here i.e. not neg
                            # perhaps add tube selection as column?
                        else:
                            # Whoa! You just hit a rare event
                            # Counts: 85 then 170 then not 0, 1, 2, 3, 4, 5
                            # Though a low number here would be a suspicious undocumented
                            # feature/cmd... let's still treat them as counts to raise
                            # the attention of a user to file an issue.
                            n = self._get_count_data(com_str)
                            self._add_to_df(n)
                            n = self._get_count_data(com_str2)
                            self._add_to_df(n)
                            n = self._get_count_data(com_str3)
                            self._add_to_df(n)
                    else:
                        # Turns out com=85 was a count and not a command flag
                        # Need to log them as counts, then
                        n = self._get_count_data(com_str)
                        self._add_to_df(n)
                        n = self._get_count_data(com_str2)
                        self._add_to_df(n)
                else:
                    n = self._get_count_data(com_str)
                    self._add_to_df(n)
        except EOFError:
            # we hit end of file
            logger.info("End of history data")
            self._raw.close()

    def get_data(self):
        """Get parsed data."""
        if self._eof_check > 0:
            # list[:-0] is empty
            return self._data[: -self._eof_check]
        return self._data

    def get_columns(self):
        """Get column names."""
        return self._columns


class _BinFile:
    """A dummy class so raw bytes and BufferedIOBase can be treated the same."""

    def __init__(self, data):
        self.data = data
        self._i = 0
        self._len_data = len(data)

    def read(self, size):
        tmp = self.data[self._i : self._i + size]
        self._i += size
        if self._i > self._len_data:
            raise EOFError
        return tmp

    def close(self):
        pass

    def tell(self):
        """Return current stream position."""
        return self._i

    def seek(self, i):
        """
        Change stream position.

        Change the stream position to the given byte offset. The offset is
        interpreted relative to the position indicated by whence.  Values
        for whence are:

        * 0 -- start of stream (the default); offset should be zero or positive
        * 1 -- current stream position; offset may be negative
        * 2 -- end of stream; offset is usually negative

        Return the new absolute position.

        Parameters
        ----------
        i int
            Position to offset to.

        Returns
        -------

        """
        self._i = i
        return self._i



"""
This mocks a device at a lower level and allows us to test pygmc.Connection()
We just need to test one device because all we really want is to test pygmc.Connection()
mock_serial only works on Linux, so skip test if not on Linux.
"""

import sys  # noqa: I001
import pytest
from serial import Serial

# D.R.Y. import from test_gmc500_plus_device_rfc_1801
from .test_gmc500_plus_device_rfc_1801 import (  # noqa: I001
    cmd_response_map,
    device_result_map,
)
import pygmc
from pygmc.connection.const import BAUDRATES


if not sys.platform.startswith("linux"):
    # TODO: can we define mock_serial after this?
    pytest.skip("skipping tests - not running linux", allow_module_level=True)


parametrize_data = [(k, v) for k, v in device_result_map.items()]


def get_mock_dev():
    """Lazily import mock_serial to avoid ModuleError in Windows"""
    mock_serial = pytest.importorskip("mock_serial", reason="Doesn't work on Win")
    # Mock device
    mock_dev = mock_serial.MockSerial()
    mock_dev.open()
    for cmd, resp in cmd_response_map.items():
        mock_dev.stub(receive_bytes=cmd, send_bytes=resp)

    # We turn heartbeat off before many ops... need to instruct mock_dev to ignore
    mock_dev.stub(receive_bytes=b"<HEARTBEAT0>>", send_bytes=b"")

    return mock_dev


def get_mock_gc():
    mock_dev = get_mock_dev()
    mock_dev_serial = Serial(mock_dev.port)

    connection = pygmc.connection.Connection(
        port="dummy", baudrate=123, serial_connection=mock_dev_serial
    )

    gc = pygmc.devices.DeviceRFC1801(connection)

    return gc


# mock_serial only works on Linux
if not sys.platform.startswith("linux"):
    pytest.skip("skipping tests - not running linux", allow_module_level=True)
else:
    mock_dev = get_mock_dev()
    mock_dev_serial = Serial(mock_dev.port)
    gc = get_mock_gc()


@pytest.mark.parametrize("cmd,expected", parametrize_data)
def test_expected_results(cmd, expected):
    """
    Test that the method/cmd in the device actually returns the expected value in the
    device_result_map dict.
    @pytest.mark.parametrize just calls this function for each item in the dictionary.
    """
    # e.g. getattr(mock_device, 'get_cpm') --> mock_device.get_cpm
    result = getattr(gc, cmd)()
    assert result == expected
    print(f"{cmd=}")
    print(f"{expected=}")
    print(f"{result=}")


def test_main_entry():
    mock_dev = get_mock_dev()
    gc = pygmc.connect(mock_dev.port)
    assert gc.get_version()[0:7] == "GMC-500"


def test_utils():
    assert isinstance(pygmc.connection.get_gmc_usb_devices(), list)
    assert isinstance(pygmc.connection.get_all_usb_devices(), list)


def test_conn_details():
    mock_dev = get_mock_dev()
    gc = pygmc.connect(mock_dev.port)
    deets = gc.get_connection_details()
    print(f"conn deets: {deets}")


def test_close_connection():
    mock_dev = get_mock_dev()
    mock_dev_serial = Serial(mock_dev.port)

    connection = pygmc.connection.Connection(
        port="dummy", baudrate=123, serial_connection=mock_dev_serial
    )
    connection.close_connection()
    connection.close_connection()  # yes, twice


def test_str_and_repr():
    mock_dev = get_mock_dev()
    mock_dev_serial = Serial(mock_dev.port)

    connection = pygmc.connection.Connection(
        port="dummy", baudrate=123, serial_connection=mock_dev_serial
    )
    print(connection)  # test __repr__ & __str__
    # disconnect to test super().__repr__
    connection._con = None
    connection.close_connection()
    print(connection)  # test super().__repr__


def test_connection_port_method():
    mock_dev = get_mock_dev()

    pygmc.connection.Connection(port=mock_dev.port, baudrate=BAUDRATES[0])


def test_connection_bad_baudrate():
    # this just enters block with warning loggers
    mock_dev = get_mock_dev()

    pygmc.connection.Connection(port=mock_dev.port, baudrate=123)


parametrize_device_conns = [
    pygmc.GMC300,
    pygmc.GMC300S,
    pygmc.GMC300EPlus,
    pygmc.GMC320,
    pygmc.GMC320Plus,
    pygmc.GMC320PlusV5,
    pygmc.GMC500,
    pygmc.GMC500Plus,
    pygmc.GMC600,
    pygmc.GMC600Plus,
    pygmc.GMC800,
    pygmc.GMCSE,
]


@pytest.mark.parametrize("dev", parametrize_device_conns)
def test_device_connection_port_method(dev):
    dev(mock_dev.port)


@pytest.mark.parametrize("dev", parametrize_device_conns)
def test_device_connection_serial_method(dev):
    connection = pygmc.connection.Connection(port=mock_dev.port, baudrate=123)
    dev(None, connection=connection)


@pytest.mark.parametrize("dev", parametrize_device_conns)
def test_device_connection_error(dev):
    with pytest.raises(ConnectionError) as excinfo:
        print(excinfo)
        dev(None)


def test_long_response_logger():
    mock_serial = pytest.importorskip("mock_serial", reason="Doesn't work on Win")
    mock_dev = mock_serial.MockSerial()
    mock_dev.open()

    mock_dev.stub(receive_bytes=b"<CONY>>", send_bytes=b"X" * 51)
    mock_dev_serial = Serial(mock_dev.port)

    connection = pygmc.connection.Connection(
        port="dummy", baudrate=123, serial_connection=mock_dev_serial
    )

    # stupid... but can't wait_sleep=0 because it fails
    # ugh, it's gonna be a non-deterministic test, fluck!
    # passes with sleep=0.000001, pad the number for github ci actions???
    response = connection.get(cmd=b"<CONY>>", wait_sleep=0.0005)
    assert len(response) == 51

    response = connection.get_at_least(cmd=b"<CONY>>", size=51, wait_sleep=0.0005)
    assert len(response) == 51



import logging
import re
from pathlib import Path

logger = logging.getLogger("pygmc.connection.udev_rule_check")


class UDevRuleCheck:
    """Check for offending Ubuntu udev rule blocking access to the GMC."""

    def __init__(self):
        """User troubleshooting class to identify offending Ubuntu udev rules."""
        # path to do .rglob on
        self._main_udev_rules_dir = Path("/usr/lib/udev/")
        # other possible paths - perhaps future expanded udev rule check?
        self._other_udev_rules_dirs = [
            Path("/etc/udev/rules.d"),
            Path("/lib/udev/rules.d"),
            Path("/run/udev/rules.d"),
            Path("/var/run/udev/rules.d"),
        ]
        # the offending path e.g. /usr/lib/udev/rules.d/85-brltty.rules
        self._brltty_path_glob = "*brltty*"
        # the \n in front is because re.search and we want to match the line
        # if it's not commented out
        self._brltty_re = '\nENV{PRODUCT}=="1a86/7523/\\*", ENV{BRLTTY_BRAILLE_DRIVER}="bm", GOTO="brltty_usb_run"'

        self._brltty_user_msg = """
        There is a linux 'udev' rule matching a GQ GMC device that is blocking the GMC
        from being accessed. It's a popular Ubuntu bug. See:
        https://askubuntu.com/questions/1403705/dev-ttyusb0-not-present-in-ubuntu-22-04

        Solutions are listed below from easiest to more involved.
        1) Run the command `sudo apt remove brltty`
            - If you are able to see, this will remove the braille (for the blind) display
            program that Ubuntu is blindly mapping the GMC to.
        2) Find the following line and comment it out.
            - ENV{PRODUCT}=="1a86/7523/*", ENV{BRLTTY_BRAILLE_DRIVER}="bm", GOTO="brltty_usb_run"
            - Most commonly found in: /usr/lib/udev/rules.d/85-brltty.rules


        TLDR: Run cmd `sudo apt remove brltty`
        """

    def get_offending_brltty_rules(self):
        """Get offending brltty udev rules."""
        paths = self._main_udev_rules_dir.rglob(self._brltty_path_glob)
        offending_brltty_udev_paths = []

        for path in paths:
            m = re.search(self._brltty_re, path.read_text())
            if not m:
                continue  # i.e. go to next path

            logger.debug(m)
            offending_brltty_udev_paths.append(str(path))

            msg = f"Found offending brltty udev rule in path:\n{path}"
            msg += "\n\n"
            msg += self._brltty_user_msg
            print(msg)  # petty enough to additionally print?
            logger.error(msg)

        return offending_brltty_udev_paths



"""
Python interface/API for GQ GMC Geiger Counter.

github: https://github.com/Wikilicious/pygmc
pypi: https://pypi.org/project/pygmc/
readthedocs: https://pygmc.readthedocs.io/
Thomaz - 2023
"""

__version__ = "0.14.1"
__author__ = "Thomaz"
__license__ = "MIT"


import logging

from pygmc.connection import Connection, Discovery
from pygmc.connection.udev_rule_check import UDevRuleCheck
from pygmc.devices import (
    GMC300,
    GMC300S,
    GMC320,
    GMC500,
    GMC600,
    GMC800,
    GMCSE,
    GMC300EPlus,
    GMC320Plus,
    GMC320PlusV5,
    GMC500Plus,
    GMC600Plus,
)
from pygmc.devices import (
    auto_get_device_from_discovery_details as _auto_get_device_class,
)
from pygmc.history import HistoryParser

logger = logging.getLogger(__name__)


def connect(
    port=None,
    baudrate=None,
    timeout=5,
):
    """
    Connect to device.

    If all parameters are None, _auto_connect() flow is used which attempts to connect
    to all available ports.
    If ANY parameter is given; it's used to refine the search, any matches are
    considered.
    Parameters are used as an OR search.

    Parameters
    ----------
    port : str | None, optional
        Exact port (device dev path / com port) e.g. '/dev/ttyUSB0'
        If port is specified, the following kwargs are ignored: vid, pid, description,
        hardware_id.
    baudrate: int | None
        Device baudrate. Leave None to auto-detect baudrate. Only applicable when port
        is specified.
    timeout: int
        Time limit for pyserial to raise timeout.

    Raises
    ------
    ConnectionError
        Unable to connect to device.
    """
    # Difficult choice... Discovery is temporary... and is nearly always instant or
    # requires user action to dis/re-connect USB. (i.e. nothing pygmc can do)
    # So why make a long timeout? Fail fast but override Discovery timeout or fail slow
    # and use user provided timeout?
    discover = Discovery(port=port, baudrate=baudrate, timeout=timeout)
    discovered_devices = discover.get_all_devices()

    if len(discovered_devices) == 0:
        # Give user direction in case of brltty udev rule blocking GMC USB device
        brltty_udev_rule_check = UDevRuleCheck()
        # line below logs & prints info for user to resolve USB connection issue
        brltty_udev_rule_check.get_offending_brltty_rules()
        raise ConnectionError("No GMC devices found.")

    device_details = discovered_devices[0]

    logger.debug(f"Selecting: {device_details}")

    device_class = _auto_get_device_class(device_details)

    gc = device_class(
        port=device_details.port, baudrate=device_details.baudrate, timeout=timeout
    )

    return gc



import datetime
import logging
import struct
from typing import Generator, Tuple

from .device import BaseDevice

logger = logging.getLogger("pygmc.devices.rfc1201")


class DeviceRFC1201(BaseDevice):
    """Device class representing Spec RFC1201"""

    def __init__(self, connection):
        """
        Represent a GMC device.

        Can be used with:
        GMC-280, GMC-300, GMC-320

        Parameters
        ----------
        connection : pygmc.Connection
            An connection interface to the USB device.
        """
        super().__init__(connection)

        # Overwrites from BaseConfig & adds 1801 specific items
        self._cfg_spec_map.update(
            {
                "Calibration_uSv_0": {
                    "index": 10,
                    "size": 4,
                    "description": "",
                    "type": "<f",
                },
                "Calibration_uSv_1": {
                    "index": 16,
                    "size": 4,
                    "description": "",
                    "type": "<f",
                },
                "Calibration_uSv_2": {
                    "index": 22,
                    "size": 4,
                    "description": "",
                    "type": "<f",
                },
                "IdleTextState": {
                    "index": 26,
                    "size": 1,
                    "description": "??",
                    "type": None,
                },
                "AlarmValue_uSv": {
                    "index": 27,
                    "size": 4,
                    "description": "",
                    "type": "<f",
                },
                "Baudrate": {
                    "index": 57,
                    "size": 1,
                    # see https://www.gqelectronicsllc.com/forum/topic.asp?TOPIC_ID=4948
                    # reply#12
                    "description": "64=1200,160=2400,208=4800,232=9600,240=14400,"
                    "244=19200,248=28800,250=38400,252=57600,254=115200",
                    "type": None,
                },
                "Threshold_uSv": {
                    "index": 65,
                    "size": 4,
                    "description": "",
                    "type": "<f",
                },
            }
        )

    def get_cpm(self) -> int:
        """
        Get CPM counts-per-minute data.

        Returns
        -------
        int
            Counts per minute
        """
        # A 16 bit unsigned integer is returned.
        # In total 2 bytes data return from GQ GMC unit.
        # The first byte is MSB byte data and second byte is LSB byte data.
        # 	  e.g.: 00 1C     the returned CPM is 28.

        cmd = b"<GETCPM>>"
        result = self.connection.get_exact(cmd, expected=b"", size=2)
        count = struct.unpack(">H", result)[0]
        return count

    def get_usv_h(self, cpm=None) -> float:
        """
        Get µSv/h as is displayed by the device. See notes below.

        cpm: int | None
            Counts per minute to be converted to µSv/h.
            Default=None polls device for live cpm to convert.
            If cpm is provided, user input is used instead.

        Notes
        -----
        Uses device calibration config.
        GQ does not have an official source documentation for device configuration nor
        cpm-to-µSv/h formula. A revision firmware update by GQ may change the config
        bytes and is often not announced.
        Treat the output as a best effort convenience.

        Returns
        -------
        float
            µSv/h

        """
        if cpm is None:
            cpm = self.get_cpm()

        # lazily load config... i.e. don't load it until it's needed.
        if not self._config:
            self.get_config()

        if not self._usv_calibration_tuple:
            calibrations = [
                (self._config["CalibrationCPM_0"], self._config["Calibration_uSv_0"]),
                (self._config["CalibrationCPM_1"], self._config["Calibration_uSv_1"]),
                (self._config["CalibrationCPM_2"], self._config["Calibration_uSv_2"]),
            ]
            self._set_usv_calibration(calibrations)

        usv_h = None
        for calib in self._usv_calibration_tuple:
            calib_max_cpm, calib_slope, calib_intercept = calib
            if cpm <= calib_max_cpm:
                usv_h = cpm * calib_slope + calib_intercept
                # valid range calibration exit condition
                break

        # if usv is still None... extrapolate from last _usv_calibration_tuple point
        # which is implicitly the highest cpm calibration via _set_usv_calibration
        if usv_h is None:
            calib_max_cpm, calib_slope, calib_intercept = self._usv_calibration_tuple[-1]
            usv_h = cpm * calib_slope + calib_intercept

        return usv_h

    def get_gyro(self) -> Tuple[int, int, int]:
        """
        Get gyroscope data.

        No units specified in spec RFC1801 nor RFC1201 :(

        Returns
        -------
        Tuple[int, int, int]
            (X, Y, Z) gyroscope data
        """
        cmd = b"<GETGYRO>>"
        # Return: Seven bytes gyroscope data in hexdecimal:
        #   BYTE1,BYTE2,BYTE3,BYTE4,BYTE5,BYTE6,BYTE7
        # Here: BYTE1,BYTE2 are the X position data in 16 bits value.
        #   The first byte is MSB byte data and second byte is LSB byte data.
        # BYTE3,BYTE4 are the Y position data in 16 bits value.
        #   The first byte is MSB byte data and second byte is LSB byte data.
        # BYTE5,BYTE6 are the Z position data in 16 bits value.
        #   The first byte is MSB byte data and second byte is LSB byte data.
        # BYTE7 always 0xAA
        result = self.connection.get_exact(cmd, expected=b"", size=7)
        x, y, z, dummy = struct.unpack(">hhhB", result)
        return x, y, z

    def get_voltage(self) -> float:
        """
        Get device voltage.

        Returns
        -------
        float
            Device voltage in volts

        """
        cmd = b"<GETVOLT>>"
        result = self.connection.get_exact(cmd, expected=b"", size=1)
        # result example: b'*'.hex() -> '2a' -> int('2a', 16) -> 42 -> 4.2V
        result = int(result.hex(), 16) / 10
        return result

    def get_datetime(self) -> datetime.datetime:
        """
        Get device datetime.

        Returns
        -------
        datetime.datetime
            Device datetime
        """
        # Return: Seven bytes data: YY MM DD HH MM SS 0xAA
        cmd = b"<GETDATETIME>>"
        data = self.connection.get_exact(cmd, expected=b"", size=7)
        year = int("20{0:2d}".format(data[0]))
        month = int("{0:2d}".format(data[1]))
        day = int("{0:2d}".format(data[2]))
        hour = int("{0:2d}".format(data[3]))
        minute = int("{0:2d}".format(data[4]))
        second = int("{0:2d}".format(data[5]))
        return datetime.datetime(year, month, day, hour, minute, second)

    def get_config(self) -> dict:
        """
        Get device config.

        Returns
        -------
        dict

        """
        cmd = b"<GETCFG>>"
        self.connection.reset_buffers()
        cfg_bytes = self.connection.get_exact(cmd, expected=b"", size=256)
        self._parse_cfg(cfg_bytes)
        return self._config

    def get_temp(self) -> float:
        """
        Get device temperature in Celsius.

        Notes
        -----
        Seems a bit temperamental. GMC-300S initially returned 0, then 86, then
        stabilized to 20.7 (which is about the room temperature when measured).
        i.e. call the method a few times until it stabilizes.

        Returns
        -------
        float
            Device temperature is celsius.

        """
        result = self.connection.get_exact(b"<GETTEMP>>", size=4)
        sign = 1
        if result[2] != 0:
            sign = -1
        temp = sign * float("{}.{}".format(result[0], result[1]))
        return temp

    def heartbeat_live(self, count=60) -> Generator[int, None, None]:
        """
        Get live CPS data, as a generator. i.e. yield (return) CPS as available.

        Parameters
        ----------
        count : int, optional
            How many CPS counts to return (default=60). Theoretically, 1count = 1second.
            Wall-clock time can be a bit higher or lower.

        Yields
        ------
        int
            CPS
        """
        self.connection.reset_buffers()
        try:
            self._heartbeat_on()
            for i in range(count):
                raw = self.connection.read_until(expected=b"", size=2)
                # only first 14 bits are used, because why not complicate things
                cps = struct.unpack(">H", raw)[0] & 0x3FFF
                yield cps
        finally:
            self._heartbeat_off()

    def heartbeat_live_print(self, count=60) -> None:
        """
        Print live CPS data.

        Parameters
        ----------
        count : int, optional
            How many CPS counts to return (default=60). Theoretically, 1count = 1second.
            Wall-clock time can be a bit higher or lower.

        """
        max_ = 0
        total = 0
        for i, cps in enumerate(self.heartbeat_live(count=count)):
            if cps > max_:
                max_ = cps
            total += cps
            # more than 4 digits for cps is scary... but let's stick to specs
            # From https://www.gqelectronicsllc.com/support/GMC_Selection_Guide.htm
            # highest cpm value is "999,999" (a strange number, unlike 65,535)
            # perhaps due to screen width limitations. let's use same for cps
            # 10 spaces (with commas) per sec... like 3 years
            # total uses 11 spaces... if you're doing more than 10 cps for three years
            # ...

            # Why the leading empty space?
            # Because the cursor blinker takes up one space and blocks view.
            msg = f" cps={cps:<7,} | max={max_:<7,} | total={total:<11,} | loop={i:<10,}"
            print(msg, end="\r")  # Carriage return - update line we just printed
        # Why extra print \n? Because \r at end causes prompt to end-up shifted at end
        print("", end="\n")  # empty print to move carriage return to next line

    def power_off(self) -> None:
        """Power OFF device."""
        cmd = b"<POWEROFF>>"
        self.connection.reset_buffers()
        self.connection.write(cmd)

    def power_on(self) -> None:
        """Power ON device."""
        cmd = b"<POWERON>>"
        self.connection.reset_buffers()
        self.connection.write(cmd)

    def send_key(self, key_number) -> None:
        """
        Send key press signal to device.

        Note the power button acts as menu clicks and does not power on/off.

        Parameters
        ----------
        key_number: int
            Each number represents a key-press.
            key=0 -> S1 (back button)
            key=1 -> S2 (down button)
            key=2 -> S3 (up button)
            key=3 -> S4 (power button)

        """
        if key_number not in (0, 1, 2, 3):
            raise ValueError("key must be in (0, 1, 2, 3)")

        cmd = "<KEY{}>>".format(key_number).encode()
        self.connection.write(cmd)

    def set_datetime(self, datetime_=None) -> None:
        """
        Set datetime on device.

        Parameters
        ----------
        datetime_: None | datetime.datetime
            Datetime to set. Default=None uses current time on computer i.e.
            datetime.datetime.now()

        Raises
        ------
        ValueError
            Year value earlier than 2000.

        RuntimeError
            Unexpected response from device.

        """
        if not datetime_:
            datetime_ = datetime.datetime.now()

        if datetime_.year < 2000 or datetime_.year >= 3000:
            # welp... device has year hardcoded 20xx
            raise ValueError("Device can't set year earlier than 2000")

        dt_cmd = struct.pack(
            ">BBBBBB",
            datetime_.year - 2000,
            datetime_.month,
            datetime_.day,
            datetime_.hour,
            datetime_.minute,
            datetime_.second,
        )
        cmd = b"<SETDATETIME" + dt_cmd + b">>"
        self.connection.reset_buffers()
        result = self.connection.get_exact(cmd, expected=b"", size=1)
        if not result == b"\xaa":
            raise RuntimeError("Unexpected response: {}".format(result))

    def reboot(self) -> None:
        """
        Reboot device.

        Note: Different from power off-on as it changes display to default.
        """
        cmd = b"<REBOOT>>"
        self.connection.write(cmd)



import logging
import re

# pypi
from serial.tools import list_ports as serial_list_ports

logger = logging.getLogger("pygmc.connection")


def get_all_usb_devices(regexp=None, include_links=True) -> list:
    """
    Get all available USB devices.

    Parameters
    ----------
    regexp : None | str, optional
        Search for ports using a regular expression. Port name, description and
        hardware ID are searched.
        hardwareID example ('USB VID:PID=1A86:7523 LOCATION=2-1')
        Default=None, find all.
    include_links : bool, optional
        include symlinks under /dev when they point to a serial port, by default True

    Returns
    -------
    list
        available ports, type [serial.tools.list_ports_linux.SysFS]
    """
    logger.debug(
        f"_get_available_usb_devices(regexp={regexp}, include_links={include_links})"
    )
    if not regexp:
        ports = serial_list_ports.comports(include_links=include_links)
    else:
        # cast as list because it's a generator and I want an easy return type
        # How many USB devices could a user possibly have?
        ports = list(serial_list_ports.grep(regexp=regexp, include_links=include_links))

    logger.debug(f"All ports/dev-devices found: {[(x.device, x.hwid) for x in ports]}")

    return ports


def get_gmc_usb_devices(
    vid="1A86", pid="7523", description=".*", include_links=True
) -> list:
    """
    Get a list of GMC USB devices.

    Default parameters are best known parameters that
    matches GMC USB devices.

    Parameters
    ----------
    vid: str | int
        Hexadecimal string or int to exact match the Vendor-ID vid.
        e.g. '0x1A86' or '1A86' or 6790 will all match.
    pid: str | int
        Hexadecimal string or int to exact match the Product-ID pid.
        e.g. '0x7523' or '7523' or 29987 will all match. (Note '7523' is considered a hex
        as a string)
    description: str
        Regex to match usb device description. (case-insensitive flag used)
        Known descriptions: 'USB2.0-Serial' & 'USB-Serial'
    include_links : bool
        Include symlinks under /dev when they point to a serial port, by default True

    Returns
    -------
    list
        List of GMC devices.

    """
    # vid = "1A86"  # 6790
    # pid = "7523"  # 29987

    if isinstance(vid, str):
        vid = int(vid, base=16)

    if isinstance(pid, str):
        pid = int(pid, base=16)

    logger.debug(
        f"get_available_usb_devices({vid}, {pid}, {description}, {include_links}"
    )

    all_devices = serial_list_ports.comports(include_links=include_links)
    logger.debug(f"All devices: {[(x.device, x.hwid) for x in all_devices]}")
    matched_devices = []
    for device in all_devices:
        vid_m = vid == device.vid
        pid_m = pid == device.pid
        description_m = re.match(
            description, str(device.description), flags=re.IGNORECASE
        )
        if all([vid_m, pid_m, description_m]):
            matched_devices.append(device)
    logger.debug(f"Matched devices: {[(x.device, x.hwid) for x in matched_devices]}")

    return matched_devices



import argparse
from pathlib import Path
from unittest.mock import mock_open, patch

import pytest

import pygmc
from pygmc import cli

from .data import data_gmc800, data_history_parser
from .mocks import MockConnection


def _get_gc_for_save_hist_tests():
    # From: test_device_get_history.py  (D.R.Y. be dammed)
    # How to create cmd input for test:
    # start_s = struct.pack(">I", 0)[1:]  # 0=start position
    # size_s = struct.pack(">H", 110)  # 110=end position
    # cmd = b"<SPIR" + start_s + size_s + b">>"
    # and set flash-mem read size to 110
    cmd_response_map = {
        b"<SPIR\x00\x00\x00\x00n>>": data_history_parser.raw_history_with_notes1,
        b"<SPIR\x00\x00\x00\x00\xd3>>": data_history_parser.raw_history_with_notes2,
        b"<SPIR\x00\x00\x00\x00\n>>": b"\xff" * 10,
        # break reading when page is all \xff
    }

    # Mock the connection to a device
    # Uses recorded responses device
    mock_connection = MockConnection(cmd_response_map)

    mock_device = pygmc.devices.BaseDevice(mock_connection)
    # Override mem/read size
    mock_device._flash_memory_size_bytes = 110
    mock_device._flash_memory_page_size_bytes = 110

    return mock_device


def _get_gc_for_live_tests():
    cmd_response_map = data_gmc800.cmd_response_map.copy()
    # This one is tricky... can only test count=1
    cmd_response_map[b"<HEARTBEAT1>>"] = b"\x00\x00\x00'"  # 39
    mock_connection = MockConnection(cmd_response_map)
    mock_device = pygmc.devices.GMC800(None, connection=mock_connection)
    return mock_device


def test_help_command(capsys):
    """GRC check - this actually checks a lot!"""
    help_msg = "usage: cli.py [-h] [-p PORT] [-b BAUDRATE] {usb,live,save}"
    with pytest.raises(SystemExit):
        cli.main(["--help"])
        captured = capsys.readouterr()
        assert "usage: cli.py" in captured.out  # simpleton check
        assert help_msg in captured.out


def test_name_main(capsys):
    """GRC check - this actually checks a lot!"""
    help_msg = "usage: cli.py [-h] [-p PORT] [-b BAUDRATE] {usb,live,save}"
    with pytest.raises(SystemExit):
        cli.main(["--help"])
        captured = capsys.readouterr()
        assert "usage: cli.py" in captured.out  # simpleton check
        assert help_msg in captured.out


@patch("pathlib.Path.exists", raises=FileExistsError)
def test_save_raises_exception_on_file_exists(mock_file):
    """Easier to change a file name than to recover a replaced file"""
    with pytest.raises(FileExistsError):
        cli.main(["save", "-f", "file_already_exists.csv"])


@patch(
    "argparse.ArgumentParser.parse_args",
    return_value=argparse.Namespace(
        port=None,
        baudrate=None,
        actions="save",
        file_name=Path("hist.csv"),
        raw=False,
    ),
)
def test_save_history(mock_args):
    mock_device = _get_gc_for_save_hist_tests()

    patcher_gc = patch("pygmc.cli._get_gc", return_value=mock_device)
    patcher_gc.start()

    with patch("builtins.open", mock_open()) as mock_file:
        cli.main()
        print(mock_file.mock_calls)
        # use position to test hist was saved correctly
        # [call(PosixPath('hist.csv'), 'w', newline=''),
        #  call().__enter__(),
        #  call().write('datetime,count,unit,mode,reference_datetime,notes\r\n'),
        #  call().write(
        #      '2020-07-26 12:45:55,66,CPM,every minute,2020-07-26 12:44:55,\r\n'),
        # ...]

        mock_file.assert_called_once_with(Path("hist.csv"), "w", newline="")

        header = mock_file.mock_calls[2].args[0]
        expected_header = "datetime,count,unit,mode,reference_datetime,notes\r\n"
        assert header == expected_header

        has_a_notes_entry = False
        tidy_count = len(data_history_parser.raw_history_with_notes1_tidy)
        for i in range(tidy_count):
            call = mock_file.mock_calls[i + 3]  # +3 to offset open, enter, header calls
            # eg '2020-07-26 13:01:26,63,CPM,every minute,2020-07-26 13:00:26,&5ABC\r\n'
            data_list = call.args[0].split(",")
            data_count = int(data_list[1])

            tidy = data_history_parser.raw_history_with_notes1_tidy[i]
            expected_count = tidy[1]
            assert data_count == expected_count, print(f"{i=}")

            if tidy[5] is not None:
                # e.g. '&5ABC\r\n'
                data_notes = data_list[5].strip()  # remove return carriage & new line
                assert tidy[5] == data_notes, print(f"{i=}")
                has_a_notes_entry = True  # cross-test

        assert has_a_notes_entry, print("Was expecting at least one notes entry")

    patcher_gc.stop()


@patch(
    "argparse.ArgumentParser.parse_args",
    return_value=argparse.Namespace(
        port=None,
        baudrate=None,
        actions="save",
        file_name=Path("hist.bin"),
        raw=True,
    ),
)
def test_save_raw_history(mock_args):
    mock_device = _get_gc_for_save_hist_tests()

    patcher_gc = patch("pygmc.cli._get_gc", return_value=mock_device)
    patcher_gc.start()

    with patch("builtins.open", mock_open()) as mock_file:
        cli.main()
        print(mock_file.mock_calls)
        mock_file.assert_called_once_with(Path("hist.bin"), "wb")
        data_write = mock_file.mock_calls[2].args[0]
        assert data_write == data_history_parser.raw_history_with_notes1

    patcher_gc.stop()


@patch(
    "argparse.ArgumentParser.parse_args",
    return_value=argparse.Namespace(
        port=None,
        baudrate=None,
        actions="live",
        time=1,
    ),
)
def test_live(mock_args, capsys):
    # This test could be improved a bit
    mock_device = _get_gc_for_live_tests()

    patcher_gc = patch("pygmc.cli._get_gc", return_value=mock_device)
    patcher_gc.start()

    cli.main()

    captured = capsys.readouterr()
    assert "GMC-800" in captured.out  # the print ver
    assert "39" in captured.out  # the CPS value

    patcher_gc.stop()



"""Discover GMC devices."""

import logging
import re
from collections import namedtuple

from serial import SerialException

from .connection import Connection
from .const import BAUDRATES
from .utils import get_gmc_usb_devices

logger = logging.getLogger("pygmc.discovery")


# Let's maintain "version" as full output from .get_version()
# Add new fields "model" & "revision"
DeviceDetails = namedtuple(
    "Device", ["port", "baudrate", "version", "serial_number", "model", "revision"]
)


class Discovery:
    """Discover GMC Devices"""

    def __init__(self, port=None, baudrate=None, timeout=3):
        """
        Discover GMC devices.

        Parameters
        ----------
        port: str | None
            Dev device, port, com, if known. Leave None to auto-discover all ports.
        baudrate: int | None
            Device baudrate, if known. Leave None to auto-discover correct baudrate.
        """
        self._timeout = timeout
        self._discovered_devices = []
        self._discover_devices_flow(port=port, baudrate=baudrate)

    def _discover_devices_flow(self, port=None, baudrate=None) -> None:
        # reset discovered devices
        self._discovered_devices = []
        if port and baudrate:
            # what are you doing in discovery?
            self._validate_device(port, baudrate)
        elif port and not baudrate:
            # likely the 2nd most used
            self._discover_with_known_port(port=port)
        elif baudrate and not port:
            # Edge case...
            self._discover_with_known_baudrate(baudrate=baudrate)
        else:
            # lemme just do everything for you... likely most common case
            self._discover_all()

    @staticmethod
    def _get_model_rev_from_version(ver: str) -> tuple:
        """
        Get model & rev from device version info.

        Parameters
        ----------
        ver: str
            Device version from get_version().

        Returns
        -------
        tuple
            (model, revision)
            Returns empty str for each if unable to match.

        """
        # See: https://www.gqelectronicsllc.com/forum/topic.asp?TOPIC_ID=10608

        # ver=GMC-800Re1.08    --> model=GMC-800    re=1.08
        # ver=GMC-600+Re 2.52  --> model=GMC-600+   re=2.52
        # ver=GMC-500+Re 2.22  --> model=GMC-500+   re=2.22
        # ver=GMC-320Re 4.26   --> model=GMC-320    re=4.26
        # ver=GMC-300SRe 1.14  --> model=GMC-300S   re=1.14
        # ver=GMC-SE Re 1.05   --> model=GMC-SE     re=1.05
        # ver=GMC 404 Re3.14A  --> model=GMC 404    re=3.14A

        # There could be a space in model (want to include)
        # and a space after the model (don't want included)
        # The '.*?' creates a non-greedy match.
        m = re.search("(.*?) ?Re ?(.*)", ver)
        if m:
            return m.groups()
        return "", ""

    @staticmethod
    def _get_gmc_usb_ports() -> list:
        devices = get_gmc_usb_devices()
        ports = [x.device for x in devices]
        # make the list intuitive
        ports = sorted(ports)
        return ports

    @staticmethod
    def _get_device_info(conn):
        if conn._con.in_waiting != 0:
            # don't intrude
            logger.debug(f"Device has non-zero bytes waiting in in-buffer ({conn._con})")
            return None
        # Device likely has heartbeat on
        if conn._con.out_waiting != 0:
            # don't intrude
            logger.debug(f"Device has non-zero bytes waiting in out-buffer ({conn._con})")
            return None

        try:
            conn.reset_buffers()
            serial_number = conn.get_exact(b"<GETSERIAL>>", size=7).hex()
            version = conn.get_at_least(b"<GETVER>>", size=7, wait_sleep=0.01).decode()
            if serial_number and version:
                return {"serial_number": serial_number, "version": version}
            else:
                return None
        except Exception as e:
            # Unsure of exception types.
            logger.warning(f"{e}", exc_info=True)
            return None
        finally:
            conn.reset_buffers()

    def _validate_device(self, port, baudrate) -> bool:
        logger.debug(f"Checking port={port} baudrate={baudrate}")
        try:
            conn = Connection(port=port, baudrate=baudrate, timeout=self._timeout)
        except SerialException as e:
            # Should a convenience method log as warning/error?
            # If only GQ Electronics would put the version/serial in USB description...
            logger.warning(f"Skipping error connecting: {e}", exc_info=True)
            return False

        info = self._get_device_info(conn)
        conn.close_connection()
        if info:
            version = info["version"]
            serial_number = info["serial_number"]
            model, revision = self._get_model_rev_from_version(version)
            discovered_device = DeviceDetails(
                port=port,
                baudrate=baudrate,
                version=version,
                model=model,
                revision=revision,
                serial_number=serial_number,
            )
            logger.debug(f"Discovered device: {discovered_device}")
            self._discovered_devices.append(discovered_device)
            return True

        return False

    def _discover_with_known_baudrate(self, baudrate) -> None:
        ports = self._get_gmc_usb_ports()
        for port in ports:
            self._validate_device(port=port, baudrate=baudrate)

    def _discover_with_known_port(self, port) -> None:
        for baudrate in BAUDRATES:
            is_valid = self._validate_device(port=port, baudrate=baudrate)
            if is_valid:
                break

    def _discover_all(self) -> None:
        ports = self._get_gmc_usb_ports()
        for port in ports:
            # D.R.Y. (don't repeat yourself)
            self._discover_with_known_port(port=port)

    def get_device_by_serial_number(self, serial_number) -> list:
        """
        Get devices by matching serial number.

        May return more than one device if sym-links are used.

        Parameters
        ----------
        serial_number: str
            Device serial number.

        Returns
        -------
        list
            List of matching devices.

        """
        # may return more than one device if user made multiple links
        matched_devices = []
        for device in self._discovered_devices:
            if device.serial_number == serial_number:
                matched_devices.append(device)
        return matched_devices

    def get_device_by_version(self, version) -> list:
        """
        Get devices by matching version.

        Parameters
        ----------
        version: str
            Device version regex. e.g. 'GMC-5' will match any device starting with GMC-5.

        Returns
        -------
        list
            List of matching devices.

        """
        matched_devices = []
        for device in self._discovered_devices:
            if re.match(version, str(device.version), flags=re.IGNORECASE):
                matched_devices.append(device)
        return matched_devices

    def get_all_devices(self) -> list:
        """
        Get all discovered devices.

        Returns
        -------
        list
            List of all discovered devices.

        """
        return self._discovered_devices



"""
Represent a USB connection to a GMC.

This is the communication class.
"""

import inspect
import logging
import time

# pypi
import serial

from .const import BAUDRATES

logger = logging.getLogger("pygmc.connection")


class Connection:
    """
    Represent a connection to a GMC device.

    A wrapper around pyserial with common operations for a GMC device.
    """

    def __init__(self, port, baudrate, timeout=5, serial_connection=None):
        """
        Represent a connection to a GMC device.

        Parameters
        ----------
        port: str
            Dev device, port, com to connect to.
            On linux, it's usually /dev/ttyUSB0 and on windows, it's usually COM3.
        baudrate: int
            Speed of communication over serial USB. Must be a compatible value.
        timeout : int, optional
            Serial connection timeout, seconds, by default 5
        serial_connection: serial.Serial | None
            An initialized Serial instance.
        """
        # pyserial has a breaking change from 3.4 to 3.5
        # TypeError:
        #     SerialBase.read_until() got an unexpected keyword argument 'expected'
        # 'terminator' for serial==3.4, 'expected' for serial==3.5
        # Doing ape logic below to resolve pyserial smooth brain breaking change
        try:
            s = inspect.signature(serial.SerialBase.read_until)
            params = list(s.parameters.keys())  # e.g. ['self', 'terminator', 'size']
            self._read_until_param_name = params[
                1
            ]  # 'terminator' for serial==3.4, 'expected' for serial==3.5
        except Exception as e:  # noqa
            logger.exception("Unable to resolve read_until param name")
            self._read_until_param_name = "expected"  # just guess

        # The connection
        if serial_connection:
            logger.debug("User provided serial connection.")
            self._con = serial_connection
        else:
            if baudrate not in BAUDRATES:
                logger.warning(f"Baudrate={baudrate} not in known compatible rates.")
                logger.warning(f"Known compatible baudrates={BAUDRATES}")
            try:
                self._con = serial.Serial(port=port, baudrate=baudrate, timeout=timeout)
            except TimeoutError:
                # See issue #112
                # Windows-running-VMware-running-Ubuntu - sleep-wake - USB issue
                # And # 118
                # Bad USB cable/connection causing unreliable communication.
                logger.exception(
                    "TimeoutError - consider reconnecting USB or replacing USB cable."
                )
                raise

    def __repr__(self):
        """Use pyserial __repr__"""
        if self._con:
            return self._con.__repr__()
        return super().__repr__()

    def __str__(self):
        """Use pyserial __repr__"""
        # maybe a nicer str?
        return str(self.__repr__())

    def get_connection_details(self) -> dict:
        """
        Get connection details.

        Values of None means not available or not applicable.

        Returns
        -------
        dict

        """
        deets = {
            "port": None,
            "baudrate": None,
            "is_open": None,
            "in_waiting": None,
            "out_waiting": None,
            "timeout": None,
        }

        for key in list(deets):
            if hasattr(self._con, key):
                deets[key] = getattr(self._con, key)

        return deets

    def close_connection(self) -> None:
        """Close connection."""
        if self._con is None:
            pass
        else:
            logger.info(f"Close connection: {self._con}")
            self._con.close()

    def reset_buffers(self) -> None:
        """
        Reset input & output buffers on pyserial connection.

        reset_input_buffer(): Clear input buffer, discarding all that is in the buffer.
        reset_output_buffer(): Clear output buffer, aborting the current output and
        discarding all that is in the buffer.
        """
        # Clear input buffer, discarding all that is in the buffer.
        logger.debug("reset_input_buffer")
        self._con.reset_input_buffer()
        # Clear output buffer,
        # aborting the current output and discarding all that is in the buffer.
        logger.debug("reset_output_buffer")
        self._con.reset_output_buffer()

    def write(self, cmd: bytes, log: bool = True) -> None:
        """
        Write command to device.

        Parameters
        ----------
        cmd : bytes
            Write command e.g. <GETVER>>
        log : bool
            Default=True to log cmd at debug level. Set false when writing sensitive
            information such as WiFi password.
        """
        if log:
            logger.debug(f"write='{cmd}'")
        else:
            logger.debug("writing cmd")
        self._con.write(cmd)
        self._con.flush()

    def read(self, wait_sleep=0.3) -> bytes:
        """
        Read all available data.

        Which may be incomplete. (noob/newbie method)

        Parameters
        ----------
        wait_sleep : float, optional
            Time to sleep to give device time to write, by default 0.3

        Returns
        -------
        bytes
            Device response
        """
        # return everything currently in device buffer
        # i.e. may be incomplete so wait a bit before read
        time.sleep(wait_sleep)
        # in pyserial==3.5 method added .read_all()
        # Read all bytes currently available in the buffer of the OS.
        # BUT... not available in pyserial==3.4
        # ADDITIONALLY, https://pyserial.readthedocs.io/en/latest/index.html
        # says latest yet refers to 3.4
        # SO... lets make this requirement 3.4 and manually implement read_all()
        if hasattr(self._con, "read_all"):
            logger.debug("read_all")
            result = self._con.read_all()
        else:
            # in_waiting - Return the number of bytes currently in the input buffer.
            logger.debug("read(in_waiting)")
            result = self._con.read(self._con.in_waiting)

        if len(result) <= 50:
            logger.debug(f"response={result}")
        else:
            # reading history data pollutes the logs with T.M.I.
            # seeing response is as easy as: logging.basicConfig(level=9)
            msg = f"response-len={len(result)} (set log level=9 to log full response)"
            logger.debug(msg)
            logger.log(level=9, msg=f"response={result}")

        return result

    def read_until(self, size=None, expected=b"") -> bytes:
        r"""
        Read device data until expected LF is reached or expected result size is reached.

        Waits until conditions met or timeout.
        Some data has '\n' which causes reading to stop.
        default changed from b'\n' to b''

        Parameters
        ----------
        size : None | int, optional
            Length of expected bytes, by default None
        expected : bytes, optional
            Expected end character, by default b''

        Returns
        -------
        bytes
            Device response
        """
        logger.debug(f"read_until(size={size}, expected={expected})")
        # This is to resolve pyserial breaking change. See __init__ above.
        params = {self._read_until_param_name: expected, "size": size}
        result = self._con.read_until(**params)
        if len(result) <= 50:
            logger.debug(f"response={result}")
        else:
            # reading history data pollutes the logs with T.M.I.
            # seeing response is as easy as: logging.basicConfig(level=9)
            msg = f"response-len={len(result)} (set log level=9 to log full response)"
            logger.debug(msg)
            logger.log(level=9, msg=f"response={result}")
        return result

    def read_at_least(self, size, wait_sleep=0.05) -> bytes:
        """
        Read at least <size> bytes then wait <wait_sleep> and read the buffer.

        i.e. Wait as long as needed to get at-least <size> bytes then wait <wait_sleep>
        seconds and read whatever else is ready in the buffer.

        Parameters
        ----------
        size: int
            Minimum size expected to read or timeout.
        wait_sleep: float | int
            Time to wait in seconds to check if there's anything remaining in the buffer.

        Notes
        -----
        This method resets the input & output buffers after; incase there was extra info
        that would've been added to the buffers.
        This is useful for ill-defined specs where there is no exact size prescribed and
        not waiting enough may result in empty/partial response and waiting too long is
        wasteful if the response was ready quickly.

        Returns
        -------
        bytes

        """
        logger.debug(f"read_at_least(size={size}, wait_sleep={wait_sleep})")

        # read until size or timeout
        min_size_result = self.read_until(size=size, expected=b"")
        extra_result = self.read(wait_sleep=wait_sleep)
        # add up results like str math
        result = min_size_result + extra_result

        if len(result) <= 50:
            logger.debug(f"combined-response={result}")
        else:
            msg = f"combined-response-len={len(result)} "
            msg += "(set log level=9 to log full response)"
            logger.debug(msg)
            logger.log(level=9, msg=f"combined-response={result}")
        self.reset_buffers()

        return result

    def get(self, cmd, wait_sleep=0.3) -> bytes:
        """
        Write command to device and get response.

        Only use in development/learning environment.
        May give incomplete/empty response if device is busy.

        Parameters
        ----------
        cmd : bytes
            Write command e.g. <GETVER>>
        wait_sleep : float, optional
            Time to sleep to give device time to write, by default 0.3

        Returns
        -------
        bytes
            Device response
        """
        logger.debug(f"get(cmd={cmd}, wait_sleep={wait_sleep})")
        self.write(cmd)
        result = self.read(wait_sleep=wait_sleep)
        return result

    def get_at_least(self, cmd: bytes, size: int, wait_sleep=0.05) -> bytes:
        """
        Write cmd, read at least <size> bytes then wait <wait_sleep> and read the buffer.

        Parameters
        ----------
        cmd : bytes
            Write command e.g. <GETVER>>
        size: int
            Minimum size expected to read or timeout.
        wait_sleep : float, optional
            Time to sleep (seconds) to give device time to write, by default 0.05

        Notes
        -----
        The reason for this method is due to the unspecified expected length in GETVER.
        We write the command then we must wait for the device to write the response.
        For GETVER on GMC300S:
        .get() wait=0.1s, two-hundred loops took 20.4 seconds and failed 13%
        .get_at_least() wait=0.05s, two-hundred loops took 15.3 seconds and 0% failed.

        Returns
        -------
        bytes
            Device response
        """
        logger.debug(f"get(cmd={cmd}, wait_sleep={wait_sleep})")
        self.write(cmd)
        result = self.read_at_least(size=size, wait_sleep=wait_sleep)
        return result

    def get_exact(self, cmd, size=None, expected=b"") -> bytes:
        """
        Write and read exact.

        Write command to device, provide expected LF or size (bytes),
        wait until either LF, size, or timeout is reached,
        then return device response.

        Parameters
        ----------
        cmd : bytes
            Write command e.g. <GETVER>>
        size : int | None, optional
            Expected response size, by default None
        expected : bytes, optional
            Expected end char, by default b''

        Returns
        -------
        bytes
            Device response
        """
        logger.debug(f"get_exact(cmd={cmd}, size={size}, expected={expected})")
        self.write(cmd)
        result = self.read_until(expected=expected, size=size)
        return result



"""
Motivation for these tests were adding kwarg 'timeout' and only GMC500Plus failed.
All devices should've failed.
Tests below capture this case.
"""

import pytest
import serial

import pygmc

parametrize_data = [
    x["device_class"] for x in pygmc.devices.auto_get_device.device_match_list
]


@pytest.mark.parametrize("device", parametrize_data)
def test_device_kwargs(device):
    keyword_arguments = {
        "port": "/dummy",
        "baudrate": 57600,
        "timeout": 5,
        "connection": None,
    }
    try:
        device(**keyword_arguments)
    except serial.serialutil.SerialException:
        # Exception example
        # serial.serialutil.SerialException: [Errno 2] could not open port /dummy:
        # [Errno 2] No such file or directory: '/dummy'
        # Test is aiming to catch "unexpected keyword argument 'timeout'" or similar
        pass



# baudrates for GMC device
# First one is most common default baudrate, second one is second most common default.
# The rest are not a default but they are an option on the device.
# Is a lower baudrate more stable?
BAUDRATES = (
    115200,
    57600,
    38400,
    28800,
    19200,
    14400,
    9600,
    4800,
    2400,
    1200,
)



import csv
import logging
import struct

from ..history import HistoryParser

logger = logging.getLogger("pygmc.device")


class BaseDevice:
    """Base device class which all devices inherit from."""

    def __init__(self, connection):
        """
        Represent a base GMC device.

        Parameters
        ----------
        connection : pygmc.Connection
            An connection interface to the USB device.
        """
        self.connection = connection

        self._flash_memory_size_bytes = 2**20  # 1 MiB
        self._flash_memory_page_size_bytes = 2**11  # 2048 B

        # the config under the hood, initialize empty and lazily create
        self._config = dict()

        # µSv calibration (max_cpm, slope, intercept)
        self._usv_calibration_tuple = tuple()

        # Best effort interpretation from:
        #     https://www.gqelectronicsllc.com/forum/topic.asp?TOPIC_ID=4948
        # self-documenting code to interpret config data
        # type=None means treat byte literally. e.g. b'\x00'[0] -> 0
        # type is a string means struct.unpack type
        self._cfg_spec_map = {
            "Power": {
                "index": 0,
                "size": 1,
                "description": "0=ON, 1=OFF... Backwards for reasons beyond comprehension.",
                "type": None,
            },
            "Alarm": {
                "index": 1,
                "size": 1,
                # Somehow this is not backwards like Power
                "description": "0=OFF, 1=ON",
                "type": None,
            },
            "Speaker": {
                "index": 2,
                "size": 1,
                "description": "0=OFF, 1=ON",
                "type": None,
            },
            "CalibrationCPM_0": {
                "index": 8,
                "size": 2,
                "description": "",
                "type": ">H",
            },
            "CalibrationCPM_1": {
                "index": 14,
                "size": 2,
                "description": "",
                "type": ">H",
            },
            "CalibrationCPM_2": {
                "index": 20,
                "size": 2,
                "description": "",
                "type": ">H",
            },
            "SaveDataType": {
                "index": 32,
                "size": 1,
                "description": "History data; 0=off, 1=CPS, 2=CPM, 3=CPM(avg/hr)",
                "type": None,
            },
            "MaxCPM": {
                "index": 49,
                "size": 2,
                "description": "MaxCPM Hi + Lo Byte",
                "type": ">H",
            },
            "Baudrate": {
                "index": 57,
                "size": 1,
                # coded differently for 300 and 500/600 series
                # see https://www.gqelectronicsllc.com/forum/topic.asp?TOPIC_ID=4948 reply#12
                "description": "see https://www.gqelectronicsllc.com/forum/topic.asp?TOPIC_ID=4948 reply#12",
                "type": None,
            },
            "BatteryType": {
                "index": 56,
                "size": 1,
                "description": "0=rechargeable, 1=non-rechargeable",
                "type": None,
            },
            "ThresholdMode": {
                "index": 64,
                "size": 1,
                "description": "0=CPM, 1=µSv/h, 2=mR/h",
                "type": None,
            },
            "ThresholdCPM": {
                "index": 62,
                "size": 2,
                "description": "",
                "type": ">H",
            },
        }

        # will likely save someone a lot of time
        # heartbeat-on keeps writing to buffer making other functionality un-parsable
        self._heartbeat_off()

        logger.debug("Initialize BaseDevice")

    def _heartbeat_off(self) -> None:
        """
        Turn heartbeat OFF.

        Stop writing data to buffer every second.
        """
        self.connection.write(b"<HEARTBEAT0>>")
        self.connection.reset_buffers()
        logger.debug("Heartbeat OFF")

    def _heartbeat_on(self) -> None:
        """
        Turn heartbeat ON.

        CPS data is automatically written to the buffer every second.
        """
        self.connection.write(b"<HEARTBEAT1>>")
        logger.debug("Heartbeat ON")

    def _read_history_position(self, start_position, chunk_size) -> bytes:
        # http://www.gqelectronicsllc.com/forum/topic.asp?TOPIC_ID=4445
        # don't need spir fix because... reset read/write buffer.
        start_s = struct.pack(">I", start_position)[1:]
        size_s = struct.pack(">H", chunk_size)

        cmd = b"<SPIR" + start_s + size_s + b">>"
        data = self.connection.get_exact(cmd, size=chunk_size)

        # MUST reset buffers... or deal with the bug
        # device with bug returns chunk_size + 1
        # if we read chunk_size, then there is one byte left in buffer
        # which will throw off all further commands
        self.connection.reset_buffers()

        return data

    def _parse_cfg(self, cfg_bytes: bytes) -> None:
        """
        Parses config bytes and sets self._config.

        Parameters
        ----------
        cfg_bytes: bytes
            Device <GETCFG>>

        Returns
        -------
        None

        """
        cfg_map = self._cfg_spec_map
        for name, d in cfg_map.items():
            i = d["index"]
            e = i + d["size"]
            raw = cfg_bytes[i:e]

            if d["type"] and d["type"] != "tbd":
                value = struct.unpack(d["type"], raw)[0]
            elif d["type"] is None:
                value = raw[0]
            else:
                logger.warning(f"config={name} not understood")
                value = None

            self._config[name] = value

    def _set_usv_calibration(self, calibrations: list) -> None:
        """
        Set µSv calibration from config.

        See https://www.gqelectronicsllc.com/forum/topic.asp?TOPIC_ID=10435
        User can convert Sievert (an SI unit of ionizing radiation) to Roentgen
        (a legacy/retired/deprecated unit of ionizing radiation)

        Parameters
        ----------
        calibrations: list
            List of tuples(int, float|int) representing max_cpm and uSv.

        Returns
        -------
        None

        """
        usv_range_slope_intercept = []  # max_cpm, slope, intercept

        # insert dummy calibration point to generalize to y=mx+b
        calibrations.insert(0, (0, 0))

        for i in range(len(calibrations) - 1):
            # *_s = start *_e = end
            calib_s = calibrations[i]
            calib_e = calibrations[i + 1]
            logger.debug(f"calib_s={calib_s} calib_e={calib_e}")

            cpm_s = calib_s[0]
            usv_s = calib_s[1]

            cpm_e = calib_e[0]
            usv_e = calib_e[1]

            dy = usv_e - usv_s
            dx = cpm_e - cpm_s

            m = dy / dx  # slope
            b = -(m * cpm_e) + usv_e  # intercept

            # The previous calib may have a max_cpm larger than the current one
            # This is actually the case in the default calibration for GMC-500+
            if cpm_e > cpm_s:
                calib = (cpm_e, m, b)
                usv_range_slope_intercept.append(calib)

        self._usv_calibration_tuple = tuple(usv_range_slope_intercept)

    def get_raw_history(self) -> bytes:
        """
        Get device history data.

        Stops reading when read entire page contains empty data.
        Full 1 MiB read takes ~5 minutes on the slower 57,600 baudrate

        Returns
        -------
        bytes
            Raw history data.

        """
        i = 0
        hist = b""
        for start_position in range(
            0, self._flash_memory_size_bytes, self._flash_memory_page_size_bytes
        ):
            data = self._read_history_position(
                start_position, self._flash_memory_page_size_bytes
            )

            if data.count(b"\xff") == self._flash_memory_page_size_bytes:
                logger.debug("Entire read block '\\xff' stop reading history")
                break
            hist += data

            i += 1
            logger.debug("Read history page {} done".format(i))

        return hist

    def save_history_raw(self, file_path) -> None:
        """
        Save raw device history to file.

        Parameters
        ----------
        file_path: str
            Path to save.

        """
        data = self.get_raw_history()
        with open(file_path, "wb") as f:
            f.write(data)

    def get_history_data(self) -> list:
        """
        Get tidy device memory history in a list of tuples.

        First row is column names.
        Columns: "datetime", "count", "unit", "mode", "reference_datetime", "notes"

        Returns
        -------
        list:
            List of tuples, first row is column names.

        """
        data = self.get_raw_history()
        h = HistoryParser(data=data)
        data = [h.get_columns()]
        data.extend(h.get_data())
        return data

    def save_history_csv(self, file_path: str) -> None:
        """
        Save device history as a CSV file.

        Parameters
        ----------
        file_path: str
            Path to save.

        """
        data = self.get_history_data()

        with open(file_path, "w", newline="") as csvfile:
            csv_writer = csv.writer(csvfile, delimiter=",")
            csv_writer.writerows(data)

    def get_version(self) -> str:
        """
        Get version of device.

        Has a sleep wait to read as spec RFC1801 doesn't specify end char nor byte size.
        i.e. SLOW.

        Returns
        -------
        str
            Device version
        """
        cmd = b"<GETVER>>"
        self.connection.reset_buffers()
        result = self.connection.get_at_least(cmd=cmd, size=7, wait_sleep=0.05)
        return result.decode("utf8")

    def get_serial(self) -> str:
        """Get serial."""
        cmd = b"<GETSERIAL>>"
        self.connection.reset_buffers()
        result = self.connection.get_exact(cmd, size=7)
        return result.hex()

    def get_connection_details(self) -> dict:
        """
        Get connection details from pyserial.

        Returns
        -------
        dict

        """
        return self.connection.get_connection_details()



import datetime

import pytest

from pygmc import devices

from .mocks import MockConnection

cmd_response_map = {
    b"<GETVER>>": b"GMC-800Re1.08",
    b"<GETSERIAL>>": b"\x03\x00H671\x06",
    b"<GETCPM>>": b"\x00\x00\t\xbf",
    b"<GETCPS>>": b"\x00\x00\x00'",
    b"<GETDATETIME>>": b"\x18\x03\t\x0f!:\xaa",
    b"<GETCFG>>": b"\x00\x00\xff\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x00\xff\xff\xff\xff\xff\x01\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x01\xff\xff\xff\xff\xff\xff\xff\x00\n\xff\xff\xff\xff\xff\x00\x01\xff\xff\xff\xff\xff\xff\xff\xff\x03\xff\x0f,\x00\x00\x06\x02\x00\x00<\x14\x00\x00x(\x00\x01,d\x00\x02X\xc8\x00\x04\xb1\x90A \x00\x00B\xc8\x00\x00CH\x00\x00C\xfa\x00\x00Dz\x00\x00D\xfa\x00\x00\x00\x00\x01\x01\x00\x00\x00d\x00\xff\x06\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff",
    b"<POWERON>>": b"",
    b"<POWEROFF>>": b"",
    b"<KEY0>>": b"",
    b"<REBOOT>>": b"",
}


device_result_map = {
    "get_cpm": 2495,
    "get_cps": 39,
    "get_datetime": datetime.datetime(2024, 3, 9, 15, 33, 58),
    "get_serial": "03004836373106",
    "get_usv_h": 16.222366710013002,
    "get_version": "GMC-800Re1.08",
    "power_on": None,
    "power_off": None,
    "reboot": None,
    "get_config": {
        "Power": 0,
        "Save_Mode": 1,
        "Power_Saving_Mode": 1,
        "LCD_Backlight_Level": 10,
        "LED": 1,
        "Fast_Estimate_Time": 3,
        "Alarm_Volume": 15,
        "Tube_Voltage": 44,
        "Calibration_CPM_1": 1538,
        "Calibration_CPM_2": 15380,
        "Calibration_CPM_3": 30760,
        "Calibration_CPM_4": 76900,
        "Calibration_CPM_5": 153800,
        "Calibration_CPM_6": 307600,
        "Calibration_USV_1": 10.0,
        "Calibration_USV_2": 100.0,
        "Calibration_USV_3": 200.0,
        "Calibration_USV_4": 500.0,
        "Calibration_USV_5": 1000.0,
        "Calibration_USV_6": 2000.0,
        "Click_Sound": 0,
        "Speaker_Volume": 0,
        "Vibration": 1,
        "Alarm_CPM": 100,
        "Theme": 0,
    },
}


parametrize_data = [(k, v) for k, v in device_result_map.items()]


# Mock the connection to a device
# Uses recorded responses device
mock_connection = MockConnection(cmd_response_map)


# Use our fake/mock connection in our real device class
mock_device = devices.DeviceSpec404(mock_connection)


@pytest.mark.parametrize("cmd,expected", parametrize_data)
def test_expected_results(cmd, expected):
    """
    Test that the method/cmd in the device actually returns the expected value in the
    device_result_map dict.
    @pytest.mark.parametrize just calls this function for each item in the dictionary.
    """
    # e.g. getattr(mock_device, 'get_cpm') --> mock_device.get_cpm
    result = getattr(mock_device, cmd)()
    assert result == expected
    print(f"{cmd=}")
    print(f"{expected=}")
    print(f"{result=}")


def test_auto_device():
    device = devices.auto_get_device_from_connection(mock_connection)
    assert isinstance(type(device), type(devices.GMC800))


def test_reset_buffers(capfd):
    mock_device.get_config()
    out, err = capfd.readouterr()
    assert "reset_buffers" in out


# uSv Tests
config_calib0 = {
    "Calibration_CPM_1": 100,
    "Calibration_CPM_2": 200,
    "Calibration_CPM_3": 300,
    "Calibration_CPM_4": 10000,
    "Calibration_CPM_5": 20000,
    "Calibration_CPM_6": 30000,
    "Calibration_USV_1": 1,
    "Calibration_USV_2": 10,
    "Calibration_USV_3": 100,
    "Calibration_USV_4": 5000,
    "Calibration_USV_5": 6000,
    "Calibration_USV_6": 7000,
}

# Tests edge case of non-increasing config cpm
config_calib1 = {
    "Calibration_CPM_1": 100,
    "Calibration_CPM_2": 25,
    "Calibration_CPM_3": 300,
    "Calibration_CPM_4": 10000,
    "Calibration_CPM_5": 20000,
    "Calibration_CPM_6": 30000,
    "Calibration_USV_1": 1,
    "Calibration_USV_2": 55,
    "Calibration_USV_3": 100,
    "Calibration_USV_4": 5000,
    "Calibration_USV_5": 6000,
    "Calibration_USV_6": 7000,
}


parametrize_calib_data = [
    # config, cpm, expected_result
    (config_calib0, 10, 0.1),
    (config_calib0, 100, 1),
    (config_calib0, 101, 1.0899999999999999),
    (config_calib0, 150, 5.5),
    (config_calib0, 200, 10),
    (config_calib0, 201, 10.900000000000006),
    (config_calib0, 250, 55),
    (config_calib0, 299, 99.10000000000002),
    (config_calib0, 300, 100),
    (config_calib0, 5000, 2474.2268041237116),
    (config_calib0, 15000, 5500.0),
    (config_calib0, 25000, 6500.0),
    (config_calib0, 30000, 7000),
    (config_calib0, 50000, 9000.0),  # tests extrapolation
    # edge cases
    (config_calib1, 25, 0.25),
    (config_calib1, 99, 0.99),
    (config_calib1, 101, 67.43636363636364),
    (config_calib1, 250, 91.81818181818181),
    (config_calib1, 300, 100),
    # # more edge cases
    # (config_calib2, 50, 0.32499998807907104),
    # (config_calib2, 1210, 7.864999977043251),
    # (config_calib2, 20000, 129.99999999202615),
    # (config_calib2, 30000, 195),
    # (config_calib2, 35000, 227.50000000398694),
]


@pytest.mark.parametrize("calib_config,cpm,expected", parametrize_calib_data)
def test_usv(calib_config, cpm, expected):
    cmd_response_map_usv = {}
    mock_connection_usv = MockConnection(cmd_response_map_usv)

    # Use our fake/mock connection in our real device class
    mock_device_usv = devices.DeviceSpec404(mock_connection_usv)

    mock_device_usv._config = calib_config

    print(f"{cpm=} {expected=} | {calib_config=}")

    mock_device_usv.get_cpm = lambda: cpm
    usv = mock_device_usv.get_usv_h()
    usv_from_input = mock_device_usv.get_usv_h(cpm=cpm)

    assert usv == expected, "uSv does not match expected config calibration"
    assert usv == usv_from_input, "manual input cpm does not match equivalent get_cpm"



import datetime
import getpass
import logging
import struct
from typing import Generator, Tuple

from .device import BaseDevice

logger = logging.getLogger("pygmc.devices.rfc1801")


class DeviceRFC1801(BaseDevice):
    """Device class representing Spec RFC1801"""

    def __init__(self, connection):
        """
        Represent a GMC device.

        Can be used with:
        GMC-500, GMC-500+, GMC-600, GMC-600+

        Parameters
        ----------
        connection : pygmc.Connection
            An connection interface to the USB device.
        """
        super().__init__(connection)

        # Overwrites from BaseConfig & adds 1801 specific items
        self._cfg_spec_map.update(
            {
                "Calibration_uSv_0": {
                    "index": 10,
                    "size": 4,
                    "description": "",
                    "type": ">f",
                },
                "Calibration_uSv_1": {
                    "index": 16,
                    "size": 4,
                    "description": "",
                    "type": ">f",
                },
                "Calibration_uSv_2": {
                    "index": 22,
                    "size": 4,
                    "description": "",
                    "type": ">f",
                },
                "IdleTextState": {
                    "index": 26,
                    "size": 1,
                    "description": "??",
                    "type": None,
                },
                "AlarmValue_uSv": {
                    "index": 27,
                    "size": 4,
                    "description": "",
                    "type": ">f",
                },
                "Baudrate": {
                    "index": 57,
                    "size": 1,
                    # see https://www.gqelectronicsllc.com/forum/topic.asp?TOPIC_ID=4948
                    # reply#14
                    "description": "0=115200, 1=1200, 2=2400, 3=4800, 4=9600, 5=14400, "
                    "6=19200, 7=28800, 8=38400, 9=57600",
                    "type": None,
                },
                "Threshold_uSv": {
                    "index": 65,
                    "size": 4,
                    "description": "",
                    "type": ">f",
                },
            }
        )

    def get_cpm(self) -> int:
        """
        Get CPM counts-per-minute data.

        Specs don't provide how CPM is computed nor if both high/low tubes are used.

        Returns
        -------
        int
            Counts per minute (GMC has 2 tubes, assumed cpm accounts for both?)
        """
        # A 32 bit unsigned integer is returned.
        # In total 4 bytes data return from GQ GMC unit.
        # The first byte is MSB byte data and fourth byte is LSB byte data.
        # e.g.: 00 00 00 1C     the returned CPM is 28. big-endian
        cmd = b"<GETCPM>>"
        result = self.connection.get_exact(cmd, expected=b"", size=4)
        count = struct.unpack(">I", result)[0]
        return count

    def get_usv_h(self, cpm=None) -> float:
        """
        Get µSv/h as is displayed by the device. See notes below.

        cpm: int | None
            Counts per minute to be converted to µSv/h.
            Default=None polls device for live cpm to convert.
            If cpm is provided, user input is used instead.

        Notes
        -----
        Uses device calibration config.
        GQ does not have an official source documentation for device configuration nor
        cpm-to-µSv/h formula. A revision firmware update by GQ may change the config
        bytes and is often not announced.
        Treat the output as a best effort convenience.

        Returns
        -------
        float
            µSv/h

        """
        if cpm is None:
            cpm = self.get_cpm()

        # lazily load config... i.e. don't load it until it's needed.
        if not self._config:
            self.get_config()

        if not self._usv_calibration_tuple:
            calibrations = [
                (self._config["CalibrationCPM_0"], self._config["Calibration_uSv_0"]),
                (self._config["CalibrationCPM_1"], self._config["Calibration_uSv_1"]),
                (self._config["CalibrationCPM_2"], self._config["Calibration_uSv_2"]),
            ]
            self._set_usv_calibration(calibrations)

        usv_h = None
        for calib in self._usv_calibration_tuple:
            calib_max_cpm, calib_slope, calib_intercept = calib
            if cpm <= calib_max_cpm:
                usv_h = cpm * calib_slope + calib_intercept
                break

        # if usv is still None... extrapolate from last _usv_calibration_tuple point
        # which is implicitly the highest cpm calibration via _set_usv_calibration
        if usv_h is None:
            calib_max_cpm, calib_slope, calib_intercept = self._usv_calibration_tuple[-1]
            usv_h = cpm * calib_slope + calib_intercept

        return usv_h

    def get_cps(self) -> int:
        """
        Get CPS counts-per-second.

        Returns
        -------
        int
            Counts per second
        """
        cmd = b"<GETCPS>>"
        result = self.connection.get_exact(cmd, expected=b"", size=4)
        count = struct.unpack(">I", result)[0]
        return count

    def get_max_cps(self) -> int:
        """
        Get the maximum counts-per-second since the device POWERED ON.

        Returns
        -------
        int
            Max counts per second observed
        """
        cmd = b"<GETMAXCPS>>"
        result = self.connection.get_exact(cmd, expected=b"", size=4)
        count = struct.unpack(">I", result)[0]
        return count

    def get_cpmh(self) -> int:
        """
        Get CPM of the high dose tube.

        Only GMC-500+ supported (spec RFC1801)

        Returns
        -------
        int
            Counts per minute on high dose tube (GMC has 2 tubes)
        """
        #
        # A 32 bit unsigned integer is returned.
        # In total 4 bytes data return from GQ GMC unit.
        # The first byte is MSB byte data and fourth byte is LSB byte data.
        # e.g.: 00 00 00 1C     the returned CPM is 28. big-endian
        cmd = b"<GETCPMH>>"
        result = self.connection.get_exact(cmd, expected=b"", size=4)
        count = struct.unpack(">I", result)[0]
        return count

    def get_cpml(self) -> int:
        """
        Get CPM of the low dose tube.

        Only GMC-500+ supported (spec RFC1801)

        Returns
        -------
        int
             Counts per minute on low dose tube (GMC has 2 tubes)
        """
        cmd = b"<GETCPML>>"
        result = self.connection.get_exact(cmd, expected=b"", size=4)
        count = struct.unpack(">I", result)[0]
        return count

    def get_datetime(self) -> datetime.datetime:
        """
        Get device datetime.

        Returns
        -------
        datetime.datetime
            Device datetime
        """
        # Return: Seven bytes data: YY MM DD HH MM SS 0xAA
        cmd = b"<GETDATETIME>>"
        data = self.connection.get_exact(cmd, expected=b"", size=7)
        year = int("20{0:2d}".format(data[0]))
        month = int("{0:2d}".format(data[1]))
        day = int("{0:2d}".format(data[2]))
        hour = int("{0:2d}".format(data[3]))
        minute = int("{0:2d}".format(data[4]))
        second = int("{0:2d}".format(data[5]))
        return datetime.datetime(year, month, day, hour, minute, second)

    def get_gyro(self) -> Tuple[int, int, int]:
        """
        Get gyroscope data.

        No units specified in spec RFC1801 nor RFC1201 :(

        Returns
        -------
        Tuple[int, int, int]
            (X, Y, Z) gyroscope data
        """
        cmd = b"<GETGYRO>>"
        # Return: Seven bytes gyroscope data in hexdecimal:
        #   BYTE1,BYTE2,BYTE3,BYTE4,BYTE5,BYTE6,BYTE7
        # Here: BYTE1,BYTE2 are the X position data in 16 bits value.
        #   The first byte is MSB byte data and second byte is LSB byte data.
        # BYTE3,BYTE4 are the Y position data in 16 bits value.
        #   The first byte is MSB byte data and second byte is LSB byte data.
        # BYTE5,BYTE6 are the Z position data in 16 bits value.
        #   The first byte is MSB byte data and second byte is LSB byte data.
        # BYTE7 always 0xAA
        result = self.connection.get_exact(cmd, expected=b"", size=7)
        x, y, z, dummy = struct.unpack(">hhhB", result)
        return x, y, z

    def get_voltage(self) -> float:
        """
        Get device voltage.

        Returns
        -------
        float
            Device voltage in volts

        Notes
        -----
        Device only has resolution to tenth of a volt despite example in spec RFC1801.

        """
        # Device only has resolution to tenth of a volt despite example in spec RFC1801.
        cmd = b"<GETVOLT>>"
        result = self.connection.get_exact(cmd, expected=b"", size=5)
        # result example: b'4.8v\x00'
        result = float(result[0:3])  # e.g. float(b'4.8')
        return result

    def get_config(self) -> dict:
        """
        Get device config.

        Returns
        -------
        dict

        """
        cmd = b"<GETCFG>>"
        self.connection.reset_buffers()
        cfg_bytes = self.connection.get_exact(cmd, expected=b"", size=512)
        self._parse_cfg(cfg_bytes)
        return self._config

    def power_off(self) -> None:
        """Power OFF device."""
        cmd = b"<POWEROFF>>"
        self.connection.reset_buffers()
        self.connection.write(cmd)

    def power_on(self) -> None:
        """Power ON device."""
        cmd = b"<POWERON>>"
        self.connection.reset_buffers()
        self.connection.write(cmd)

    def heartbeat_live(self, count=60) -> Generator[int, None, None]:
        """
        Get live CPS data, as a generator.

        i.e. yield (return) CPS as available.

        Parameters
        ----------
        count : int, optional
            How many CPS counts to return (default=60). Theoretically, 1count = 1second.
            Wall-clock time can be a bit higher or lower.

        Yields
        ------
        int
            CPS - Counts-Per-Second int

        """
        self.connection.reset_buffers()
        try:
            self._heartbeat_on()
            for i in range(count):
                raw = self.connection.read_until(expected=b"", size=4)
                cps = struct.unpack(">I", raw)[0]
                yield cps
        finally:
            self._heartbeat_off()

    def heartbeat_live_print(self, count=60) -> None:
        """
        Print live CPS data.

        Parameters
        ----------
        count : int, optional
            How many CPS counts to return (default=60). Theoretically, 1count = 1second.
            Wall-clock time can be a bit higher or lower.

        """
        max_ = 0
        total = 0
        for i, cps in enumerate(self.heartbeat_live(count=count)):
            if cps > max_:
                max_ = cps
            total += cps
            # more than 4 digits for cps is scary... but let's stick to specs
            # From https://www.gqelectronicsllc.com/support/GMC_Selection_Guide.htm
            # highest cpm value is "999,999" (a strange number, unlike 65,535)
            # perhaps due to screen width limitations. let's use same for cps
            # 10 spaces (with commas) per sec... like 3 years
            # total uses 11 spaces... if you're doing more than 10 cps for three years
            # ...

            # Why the leading empty space?
            # Because the cursor blinker takes up one space and blocks view.
            msg = f" cps={cps:<7,} | max={max_:<7,} | total={total:<11,} | loop={i:<10,}"
            print(msg, end="\r")  # Carriage return - update line we just printed
        # Why extra print \n? Because \r at end causes prompt to end-up shifted at end
        print("", end="\n")  # empty print to move carriage return to next line

    def send_key(self, key_number) -> None:
        """
        Send key press signal to device.

        Note the power button acts as menu clicks and does not power on/off.

        Parameters
        ----------
        key_number: int
            Each number represents a key-press.
            key=0 -> S1 (back button)
            key=1 -> S2 (down button)
            key=2 -> S3 (up button)
            key=3 -> S4 (power button)
        """
        if key_number not in (0, 1, 2, 3):
            raise ValueError("key must be in (0, 1, 2, 3)")

        cmd = "<KEY{}>>".format(key_number).encode()
        self.connection.write(cmd)

    def set_datetime(self, datetime_=None) -> None:
        """
        Set datetime on device.

        Parameters
        ----------
        datetime_: None | datetime.datetime
            Datetime to set. Default=None uses current time on computer i.e.
            datetime.datetime.now()

        Raises
        ------
        ValueError
            Year value earlier than 2000.

        RuntimeError
            Unexpected response from device.

        """
        if not datetime_:
            datetime_ = datetime.datetime.now()

        if datetime_.year < 2000 or datetime_.year >= 3000:
            # welp... device has year hardcoded 20xx
            raise ValueError("Device has year hardcoded 20xx can't set given year.")

        dt_cmd = struct.pack(
            ">BBBBBB",
            datetime_.year - 2000,
            datetime_.month,
            datetime_.day,
            datetime_.hour,
            datetime_.minute,
            datetime_.second,
        )
        cmd = b"<SETDATETIME" + dt_cmd + b">>"
        self.connection.reset_buffers()
        result = self.connection.get_exact(cmd, expected=b"", size=1)
        if not result == b"\xaa":
            raise RuntimeError("Unexpected response: {}".format(result))

    def reboot(self) -> None:
        """
        Reboot device.

        Note: Different from power off-on as it changes display to default.

        """
        cmd = b"<REBOOT>>"
        self.connection.write(cmd)

    def set_wifi_on(self) -> None:
        """Set WiFi On"""
        cmd = b"<WiFiON>>"
        result = self.connection.get_exact(cmd, size=1)
        if result != b"\xaa":
            raise RuntimeError("Unexpected response: {}".format(result))

    def set_wifi_off(self) -> None:
        """Set WiFi Off"""
        cmd = b"<WiFiOFF>>"
        result = self.connection.get_exact(cmd, size=1)
        if result != b"\xaa":
            raise RuntimeError("Unexpected response: {}".format(result))

    def set_wifi_ssid(self, ssid, bytes_encoding: str = "utf8") -> None:
        r"""
        Set WiFi SSID (Access point name)

        Parameters
        ----------
        ssid: str
            WiFi SSID access point name.
        bytes_encoding: str
            Encoding to cast to bytes. Realize it's not about the WiFi spec it's about
            the GQ & ESP8266 spec. (undocumented in GQ-RFC1801)

        Raises
        ------
        RuntimeError:
            GQ spec specifies a b"\xaa" response for success.
            RuntimeError for Unexpected response.

        """
        cmd = b"<SETSSID" + bytes(ssid, encoding=bytes_encoding) + b">>"
        result = self.connection.get_exact(cmd, size=1)
        if result != b"\xaa":
            raise RuntimeError("Unexpected response: {}".format(result))

    def set_wifi_password(self, password=None, bytes_encoding: str = "utf8"):
        r"""
        Set WiFi password

        Parameters
        ----------
        password: str | None
            Set WiFi password. Default=None prompts the user with interactive Python
            built-in getpass; to discretely input the variable. Optionally, the user can
            use the parameter to set the variable with their own method.
        bytes_encoding: str
            Encoding to cast to bytes. Realize it's not about the WiFi spec it's about
            the GQ & ESP8266 spec. (undocumented in GQ-RFC1801)

        Raises
        ------
        RuntimeError:
            GQ spec specifies a b"\xaa" response for success.
            RuntimeError for Unexpected response.

        """
        if password is None:
            # enter password via Python's getpass built-in library
            password = getpass.getpass()
        cmd = b"<SETWIFIPW" + bytes(password, encoding=bytes_encoding) + b">>"
        # don't log usb write command
        self.connection.write(cmd, log=False)
        # confirm success
        result = self.connection.read_until(size=1)
        if result != b"\xaa":
            raise RuntimeError("Unexpected response: {}".format(result))

    def set_gmcmap_user_id(self, user_id: str) -> None:
        r"""
        Set User-Id for gmcmap.com

        See: https://gmcmap.com/

        Parameters
        ----------
        user_id: str
            User-Id on gmcmap.com

        Raises
        ------
        RuntimeError:
            GQ spec specifies a b"\xaa" response for success.
            RuntimeError for Unexpected response.

        """
        cmd = b"<SETUSERID" + bytes(user_id, encoding="utf8") + b">>"
        result = self.connection.get_exact(cmd, size=1)
        if result != b"\xaa":
            raise RuntimeError("Unexpected response: {}".format(result))

    def set_gmcmap_counter_id(self, counter_id: str):
        r"""
        Set Counter-Id for currently connected GQ GMC for gmcmap.com

        Parameters
        ----------
        counter_id: str
            Counter-Id for gmcmap.com

        Raises
        ------
        RuntimeError:
            GQ spec specifies a b"\xaa" response for success.
            RuntimeError for Unexpected response.

        """
        cmd = b"<SETCOUNTERID" + bytes(counter_id, encoding="utf8") + b">>"
        result = self.connection.get_exact(cmd, size=1)
        if result != b"\xaa":
            raise RuntimeError("Unexpected response: {}".format(result))



"""PyGMC CLI - Command-Line-Interface"""

import argparse
import datetime
from pathlib import Path

try:
    from .connection import (
        Discovery,
        get_all_usb_devices,
        get_gmc_usb_devices,
    )
    from .connection.udev_rule_check import UDevRuleCheck
    from .devices import (
        auto_get_device_from_discovery_details as _auto_get_device_class,
    )
except ImportError:
    # Most likely error while trying 'python ./some_path/pygmc/cli.py --help'
    # ImportError: attempted relative import beyond top-level package
    # The issue is pygmc/history.py is at same level as pygmc/cli.py
    print("See documentation for correct usage.")
    raise


# Would've liked this in a function... but...
# don't know how to auto-document cli w/o parser out in the open like this... welp...
parser = argparse.ArgumentParser()

parser.add_argument(
    "-p",
    "--port",
    type=str,
    help="The USB port/com/dev e.g. /dev/ttyUSB0 Default=None will auto detect a port.",
    metavar=None,
)
parser.add_argument(
    "-b",
    "--baudrate",
    type=int,
    help="USB communication baudrate Default=None will auto detect the baudrate.",
    metavar=None,
)

subparsers = parser.add_subparsers(
    title="Actions",
    help="PyGMC CLI Actions:" "Common actions for PyGMC",
    dest="actions",
)

# ACTION - list usb's
parser_usb = subparsers.add_parser(
    "usb", help="List USB devices found - filtered for GMC devices"
)
parser_usb.add_argument(
    "--all", action="store_true", help="List all USBs i.e. unfiltered"
)

# ACTION - live cps
parser_live = subparsers.add_parser("live", help="a help")
parser_live.add_argument(
    "-t", "--time", type=int, default=10, help="Time is seconds to live print"
)

# ACTION - save history
parser_history = subparsers.add_parser(
    "save", help="Save device history. Default is a tidy csv file."
)
parser_history.add_argument(
    "-f",
    "--file-name",
    type=Path,
    required=True,
    help="File path. e.g. ~/hist.csv (save history as csv in linux home dir)",
)
parser_history.add_argument(
    "--raw", action="store_true", help="Save raw as-is/unmodified device history."
)


def _list_usb_flow(show_all=False):
    """Print simple information on USB devices."""
    if show_all:
        usbs = get_all_usb_devices()
    else:
        usbs = get_gmc_usb_devices()
    for usb in usbs:
        # would be nice to drop python3.7 support and use f"{value=}" formatting
        print(f"device={usb.device} description={usb.description} hwid={usb.hwid}")


def _get_gc(args):
    """Get geiger-counter-class with provided info from pygmc Discovery."""
    # Ugh... violating D.R.Y.
    discover = Discovery(port=args.port, baudrate=args.baudrate, timeout=5)
    discovered_devices = discover.get_all_devices()
    if len(discovered_devices) == 0:
        # Give user direction in case of brltty udev rule blocking GMC USB device
        brltty_udev_rule_check = UDevRuleCheck()
        # line below logs & prints info for user to resolve USB connection issue
        brltty_udev_rule_check.get_offending_brltty_rules()
        raise ConnectionError("No GMC devices found.")

    device_details = discovered_devices[0]
    device_class = _auto_get_device_class(device_details)
    gc = device_class(
        port=device_details.port, baudrate=device_details.baudrate, timeout=5
    )
    return gc


def _live_flow(args):
    """Print live data (heartbeat)."""
    gc = _get_gc(args)
    print("PyGMC - Live GMC Heartbeat")
    ver = gc.get_version()
    print(f"{ver}")
    datetime_str = datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S")
    print(f"Start={datetime_str} counts={args.time:,}")
    gc.heartbeat_live_print(count=args.time)

    datetime_finish_str = datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S")
    print(f"Finish={datetime_finish_str}")


def _cli_flow_save(args):
    """Save device history as tidy CSV or raw data."""
    if args.file_name.exists():
        # Rather give user an error than the alternative... (why you edit my file)
        raise FileExistsError(f"File exists: {args.file_name}")
    gc = _get_gc(args)
    if not args.raw:
        # user want's pygmc tidy data... good Lad or Lass or
        # Shklee or Shklim or Shkler (Futurama)
        gc.save_history_csv(args.file_name)
    else:
        # user is hotshot and want's it raw...
        gc.save_history_raw(args.file_name)


def main(argv=None) -> None:
    """
    CLI entry point.

    Parameters
    ----------
    argv: None | list
        Default=None for correct behavior. Main purpose is for unit-tests.

    Returns
    -------
    None

    """
    args = parser.parse_args(argv)

    if args.actions == "usb":
        _list_usb_flow(show_all=args.all)
    elif args.actions == "live":
        _live_flow(args)
    elif args.actions == "save":
        _cli_flow_save(args)


if __name__ == "__main__":
    main()  # pragma: no cover



import pygmc

from .data import data_history_parser


def test_history_parser_save_modes():
    h = pygmc.HistoryParser(data=data_history_parser.raw_history_with_save_modes)
    tidy_data = h.get_data()
    assert tidy_data == data_history_parser.raw_history_with_save_modes_tidy


def test_history_parser_note_decode_exception():
    raw_data = b"\xff" * 6 + b"\x55\xaa\x02\x01\x80"
    h = pygmc.HistoryParser(data=raw_data)
    assert h.get_data() == []


def test_history_parser_count_same_as_cmd():
    # consider improving parser to include 85 when it is last byte
    # currently it doesn't make it into tidy data

    raw_data = b"U\xaa\x00\x18\x01\x19\x15\x05\x0cU\xaa\x02\x55"
    h = pygmc.HistoryParser(data=raw_data)
    assert h.get_data() == []

    raw_data += b"\x12"  # add an 18 count to trigger adding 85 and 18
    h = pygmc.HistoryParser(data=raw_data)
    tidy_data = h.get_data()
    assert tidy_data[0][1] == 85
    assert tidy_data[1][1] == 18


def test_tube_selection():
    # purposefully as arg instead of kwarg
    h = pygmc.HistoryParser(data_history_parser.raw_history_tube_selection)
    assert h.get_data() == data_history_parser.raw_history_tube_selection_tidy


def test_three_byte_count():
    h = pygmc.HistoryParser(data_history_parser.raw_history_3byte_count)
    assert h.get_data() == data_history_parser.raw_history_3byte_count_tidy


def test_four_byte_count():
    # fake/synthetic data - not validated because I don't have a source to do so
    h = pygmc.HistoryParser(data_history_parser.raw_history_4byte_count)
    assert h.get_data() == data_history_parser.raw_history_4byte_count_tidy
